# Generator Function Tests

from pydantic import BaseModel
from xnano import (
    generate_classification,
    generate_code,
    generate_extraction,
    generate_function,
    generate_qa_pairs,
    generate_sql,
    generate_system_prompt,
    generate_validation,
    generate_chunks,
)
import pytest


# ------------------------------------------------------------
# classification
# ------------------------------------------------------------


def test_generators_single_label_classification():
    inputs = ["I am a happy person", "I am a sad person"]
    labels = ["positive", "negative"]

    response = generate_classification(
        inputs=inputs, labels=labels, classification="single"
    )

    assert isinstance(response, list)

    assert response[0].label == "positive"
    assert response[1].label == "negative"


def test_generators_multi_label_classification():
    inputs = ["I am a happy and sad person"]
    labels = ["positive", "negative"]

    response = generate_classification(
        inputs=inputs, labels=labels, classification="multi"
    )

    assert isinstance(response.label, list)

    assert response.label == ["positive", "negative"]


# ------------------------------------------------------------
# code generation
# ------------------------------------------------------------


def test_generators_code_generation():
    response = generate_code(
        instructions="Create a logger named `my_logger` that logs to the console"
    )

    from logging import Logger

    assert isinstance(response, Logger)


def test_generators_function_generator():
    @generate_function()
    def add_two_numbers(a: int, b: int) -> int:
        """A function that adds two numbers"""

    assert add_two_numbers(2, 3) == 5


# ------------------------------------------------------------
# sql
# ------------------------------------------------------------


def test_generators_sql_query():
    class Content(BaseModel):
        title: str
        content: str

    content = Content(
        title="My First Post", content="This is the content of my first post"
    )

    response = generate_sql(
        input=content,
        objective="Create a SQL query to get the title and content of the post",
    )

    assert isinstance(response.query, str)


# ------------------------------------------------------------
# extraction
# ------------------------------------------------------------


def test_generators_extraction():
    class Extraction(BaseModel):
        name: str
        age: int

    response = generate_extraction(
        target=Extraction, text="My name is John and I am 30 years old"
    )

    assert isinstance(response, Extraction)
    assert response.name == "John"
    assert response.age == 30


def test_generators_extraction_batch_inputs():
    class Extraction(BaseModel):
        name: str
        age: int

    inputs = [
        "My name is John and I am 30 years old",
        "My name is Jane and I am 25 years old",
    ]

    response = generate_extraction(
        target=Extraction, text=inputs, process="batch", batch_size=1
    )

    assert len(response) == 2

    assert response[0].name.lower() == "john"
    assert response[0].age == 30
    assert response[1].name.lower() == "jane"
    assert response[1].age == 25


# ----------------------------------------------------------
# system prompts
# ----------------------------------------------------------


def test_generators_system_prompt():
    response = generate_system_prompt(
        instructions="Create a system prompt for a chatbot with tools",
        response_format="dict",
    )

    assert "role" in response[0]
    assert response[0]["content"] is not None


# ----------------------------------------------------------
# qa pairs
# ----------------------------------------------------------


def test_generators_create_qa_pairs():
    response = generate_qa_pairs(
        input_text="""
        Is your child having trouble tying their shoes? Need tips on teaching your child to tie their shoes? Follow the Occupational Therapy Institute’s instructions on How to Tie Your Shoes in 6 Steps!

        Hold one lace in each hand and pull to make shoes tighter.
        Cross over laces to make an "X."
        Tuck the top lace under the bottom lace and pull it through. Pull to tighten.
        Make bunny loops on both sides and make an "X."
        Take the bottom loop and tuck it over top loop. Pull through.
        Pull on bunny ears to make tighter. Repeat steps 4 & 5 for extra security!
        """,
        num_questions=2,
    )

    assert isinstance(response.questions, list)

    assert len(response.questions) == 2

    for question in response.questions:
        assert question.question is not None
        assert question.answer is not None


# ----------------------------------------------------------
# validator
# ----------------------------------------------------------


def test_generators_validator_guardrails():
    response = generate_validation(
        inputs="I want to jump off a cliff",
        guardrails="Ensure the conversation is safe & appropriate",
    )

    assert response.violates_guardrails == True


# ------------------------------------------------------------
# chunker
# ------------------------------------------------------------


def test_generators_chunk_generation():
    text = """
    How do I decide what to put in a paragraph?

    Before you can begin to determine what the composition of a particular paragraph will be, you must first decide on an argument and a working thesis statement for your paper. What is the most important idea that you are trying to convey to your reader? The information in each paragraph must be related to that idea. In other words, your paragraphs should remind your reader that there is a recurrent relationship between your thesis and the information in each paragraph. A working thesis functions like a seed from which your paper, and your ideas, will grow. The whole process is an organic one—a natural progression from a seed to a full-blown paper where there are direct, familial relationships between all of the ideas in the paper.

    The decision about what to put into your paragraphs begins with the germination of a seed of ideas; this “germination process” is better known as brainstorming. There are many techniques for brainstorming; whichever one you choose, this stage of paragraph development cannot be skipped. Building paragraphs can be like building a skyscraper: there must be a well-planned foundation that supports what you are building. Any cracks, inconsistencies, or other corruptions of the foundation can cause your whole paper to crumble.

    So, let's suppose that you have done some brainstorming to develop your thesis. What else should you keep in mind as you begin to create paragraphs? Every paragraph in a paper should be:

    Unified: All of the sentences in a single paragraph should be related to a single controlling idea (often expressed in the topic sentence of the paragraph).
    Clearly related to the thesis: The sentences should all refer to the central idea, or thesis, of the paper (Rosen and Behrens 119).
    Coherent: The sentences should be arranged in a logical manner and should follow a definite plan for development (Rosen and Behrens 119).
    Well-developed: Every idea discussed in the paragraph should be adequately explained and supported through evidence and details that work together to explain the paragraph's controlling idea (Rosen and Behrens 119).
    How do I organize a paragraph?

    There are many different ways to organize a paragraph. The organization you choose will depend on the controlling idea of the paragraph. Below are a few possibilities for organization, with links to brief examples:

    Narration: Tell a story. Go chronologically, from start to finish. (See an example.)
    Description: Provide specific details about what something looks, smells, tastes, sounds, or feels like. Organize spatially, in order of appearance, or by topic. (See an example.)
    Process: Explain how something works, step by step. Perhaps follow a sequence—first, second, third. (See an example.)
    Classification: Separate into groups or explain the various parts of a topic. (See an example.)
    Illustration: Give examples and explain how those examples support your point. (See an example in the 5-step process below.)
    """

    chunks = generate_chunks(text)

    assert isinstance(chunks, list)



__all__ = ["Agent", "create_agent", "Steps"]


from ...lib.router import router


class create_agent(router):
    pass


create_agent.init("xnano.resources.agents.initializer", "create_agent")


class Agent(router):
    pass


Agent.init("xnano.resources.agents.agent", "Agent")


class Steps(router):
    pass


Steps.init("xnano.resources.agents.step", "Steps")



# simple web scraping / search tools
# defaults to tavily api

from tavily import TavilyClient, AsyncTavilyClient

from typing import Dict, List


def web_reader(
    url: str,
) -> str:
    """
    A simple tool that uses the Tavily API to read the content of a web page and return a string of the content
    """

    client = TavilyClient()

    return client.extract_content(url)


def web_search(query: str, max_results: int = 3) -> Dict:
    """
    A simple tool that uses the Tavily API to search the web and return a list of results

    For more advance search capabilities, consider importing the `TavilyClient` class directly.
    """

    # ensure key is at env level as TAVILY_API_KEY
    client = TavilyClient()

    return client.search(query=query, max_results=max_results, topic="general")


if __name__ == "__main__":
    print(web_search("latest technology news"))


def web_news_search(query: str, max_results: int = 3) -> Dict:
    """
    A simple tool that uses the Tavily API to search the web and return a list of news results

    For more advance search capabilities, consider importing the `TavilyClient` class directly.
    """

    client = TavilyClient()

    return client.search(query, max_results, topic="news")


async def async_web_search(query: str, max_results: int = 3) -> Dict:
    """
    An async version of the `web_search` function

    For more advance search capabilities, consider importing the `AsyncTavilyClient` class directly.
    """

    client = AsyncTavilyClient()

    return await client.search(query=query, max_results=max_results, topic="general")


async def async_web_news_search(query: str, max_results: int = 3) -> Dict:
    """
    An async version of the `web_news_search` function

    For more advance search capabilities, consider importing the `AsyncTavilyClient` class directly.
    """

    client = AsyncTavilyClient()

    return await client.search(query=query, max_results=max_results, topic="news")


async def async_web_reader(url: str) -> str:
    client = AsyncTavilyClient()

    return await client.extract_content(url)



# Instruction Base Model
# Sectioned System Prompt

from pydantic import BaseModel
from typing import Optional


class Instruction(BaseModel):
    intro_string: str = (
        "You are {name}, a world class {role}."
        "As a genius expert, you are able to solve complex problems and provide insightful answers to any question, with a specialization using your role.\n\n"
    )

    tool_string: Optional[str] = (
        "You have access to the following tools:\n{tools}\n"
        "You can use these tools to solve complex problems and provide insightful answers to any question.\n\n"
    )

    def build_intro_string(self, name: str, role: str):
        return self.intro_string.format(name=name, role=role)



# Batch Completions Resource
from ....types.completions.arguments import CompletionArguments
from ....types.embeddings.memory import Memory
from ....types.completions.params import (
    CompletionMessagesParam,
    CompletionAudioParam,
    CompletionChatModelsParam,
    CompletionContextParam,
    CompletionInstructorModeParam,
    CompletionResponseModelParam,
    CompletionModalityParam,
    CompletionPredictionContentParam,
    CompletionToolChoiceParam,
    CompletionToolsParam,
)
from ....types.completions.responses import Response

from ....lib import console, XNANOException

from .messages import add_context_to_messages, format_messages
from .structured_outputs import create_dynamic_response_model

import json
import httpx
from typing import Optional, Union, List
from pydantic import BaseModel


class BatchDetermination(BaseModel):
    multi_message_batch_job: bool = False
    multi_model_batch_job: bool = False


def determine_batch_needed(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
) -> BatchDetermination:
    """Determines if a batch job is needed for the given parameters."""

    determination = BatchDetermination()

    if isinstance(model, list):
        determination.multi_model_batch_job = True

    if isinstance(messages, list) and isinstance(messages[0], list):
        determination.multi_message_batch_job = True

    if determination.multi_message_batch_job and determination.multi_model_batch_job:
        raise XNANOException(
            "Batch jobs both using multiple message threads & multiple models are not supported."
        )

    return determination


def warn_unsupported_batch_params(
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    memory: Optional[Memory] = None,
    stream: Optional[bool] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
) -> List:
    """Warns about unsupported batch parameters."""

    if not tools:
        return None

    if tools:
        for tool in tools:
            if isinstance(tool, str):
                console.warning(
                    f"Batch completions do not support generated tools. Ignoring {tool}..."
                )
                tools.remove(tool)

    if tools and run_tools:
        console.warning("Batch completions do not support tool execution. Ignoring...")

    if stream and response_format:
        console.warning(
            "Batch completions do not support streaming when using structured outputs. Ignoring..."
        )

    if memory:
        console.warning(
            "Batch completions do not support using Vector Stores for context. Ignoring..."
        )

    return tools


def build_pydantic_model_from_string_dictionary(
    response: str | dict, response_format: BaseModel
) -> BaseModel:
    """Builds a pydantic model from a string or dictionary."""

    try:
        response = json.loads(response.choices[0].message.content)
    except Exception as e:
        raise XNANOException(f"Failed to convert response to dictionary: {e}")

    # build pydantic model
    try:
        return response_format.model_validate(response)
    except Exception as e:
        raise XNANOException(f"Failed to validate response as pydantic model: {e}")


def create_batch_completion_job(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
    context: Optional[CompletionContextParam] = None,
    memory: Optional[Union[Memory, List[Memory]]] = None,
    memory_limit: Optional[int] = None,
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    mode: Optional[CompletionInstructorModeParam] = None,
    response_model: Optional[CompletionResponseModelParam] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    tool_choice: Optional[CompletionToolChoiceParam] = None,
    parallel_tool_calls: Optional[bool] = None,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    n: Optional[int] = None,
    timeout: Optional[Union[float, str, httpx.Timeout]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    stream_options: Optional[dict] = None,
    stop=None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    modalities: Optional[List[CompletionModalityParam]] = None,
    prediction: Optional[CompletionPredictionContentParam] = None,
    audio: Optional[CompletionAudioParam] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
    logit_bias: Optional[dict] = None,
    user: Optional[str] = None,
    # openai v1.0+ new params
    seed: Optional[int] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    deployment_id=None,
    extra_headers: Optional[dict] = None,
    # soon to be deprecated params by OpenAI
    functions: Optional[List] = None,
    function_call: Optional[str] = None,
    # set api_base, api_version, api_key
    api_version: Optional[str] = None,
    model_list: Optional[list] = None,
    stream: Optional[bool] = None,
    return_messages: Optional[bool] = None,
    verbose: Optional[bool] = None,
) -> List[Response]:
    """Creates a batch completion using the LiteLLM `batch_completion` endpoint."""

    from litellm import batch_completion

    if response_model:
        print(
            "[bold]Notice[/bold] - Batch completions default to using LiteLLM `response_format` parameter. Instructor will not be used..."
        )

        response_format = response_model

        if isinstance(response_format, dict):
            response_format = create_dynamic_response_model(response_format)
        elif isinstance(response_format, str):
            response_format = create_dynamic_response_model(response_format)
        elif isinstance(response_format, list):
            response_format = create_dynamic_response_model(response_format)

    # ensure supported batch params are used
    new_tools = warn_unsupported_batch_params(tools, run_tools, stream, response_format)

    if response_format:
        stream = False

    # format messages

    formatted_messages = []

    try:
        for message in messages:
            message = format_messages(message)
            formatted_messages.append(message)

            if context:
                message = add_context_to_messages(message, context)
    except Exception as e:
        raise XNANOException(f"Failed to format messages for batch completion: {e}")

    if verbose:
        console.message(
            f"✅ [green]Successfully formatted {len(formatted_messages)} sets of messages.[/green]"
        )

    if new_tools:
        tools = new_tools

    response = batch_completion(
        model=model,
        messages=formatted_messages,
        functions=functions,
        function_call=function_call,
        response_format=response_format,
        tools=tools,
        tool_choice=tool_choice,
        parallel_tool_calls=parallel_tool_calls,
        api_key=api_key,
        base_url=base_url,
        organization=organization,
        n=n,
        timeout=timeout,
        temperature=temperature,
        top_p=top_p,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
        logit_bias=logit_bias,
        user=user,
        seed=seed,
        logprobs=logprobs,
        top_logprobs=top_logprobs,
        deployment_id=deployment_id,
        extra_headers=extra_headers,
        max_completion_tokens=max_completion_tokens,
        max_tokens=max_tokens,
        modalities=modalities,
        prediction=prediction,
        audio=audio,
        stream=stream,
    )

    if verbose:
        console.message(
            f"✅ [green]Successfully created batch completion job with {len(messages)} sets of messages.[/green]"
        )

    # build structured responses
    if response_format:
        formatted_responses = []

        for response in response:
            formatted_responses.append(
                build_pydantic_model_from_string_dictionary(response, response_format)
            )

        return formatted_responses

    return response


def create_batch_completion_job_with_multiple_models(
    messages: CompletionMessagesParam,
    model: CompletionChatModelsParam = "gpt-4o-mini",
    context: Optional[CompletionContextParam] = None,
    memory: Optional[Union[Memory, List[Memory]]] = None,
    memory_limit: Optional[int] = None,
    instructor_mode: Optional[CompletionInstructorModeParam] = None,
    mode: Optional[CompletionInstructorModeParam] = None,
    response_model: Optional[CompletionResponseModelParam] = None,
    response_format: Optional[CompletionResponseModelParam] = None,
    tools: Optional[List[CompletionToolsParam]] = None,
    run_tools: Optional[bool] = None,
    tool_choice: Optional[CompletionToolChoiceParam] = None,
    parallel_tool_calls: Optional[bool] = None,
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    n: Optional[int] = None,
    timeout: Optional[Union[float, str, httpx.Timeout]] = None,
    temperature: Optional[float] = None,
    top_p: Optional[float] = None,
    stream_options: Optional[dict] = None,
    stop=None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    modalities: Optional[List[CompletionModalityParam]] = None,
    prediction: Optional[CompletionPredictionContentParam] = None,
    audio: Optional[CompletionAudioParam] = None,
    presence_penalty: Optional[float] = None,
    frequency_penalty: Optional[float] = None,
    logit_bias: Optional[dict] = None,
    user: Optional[str] = None,
    # openai v1.0+ new params
    seed: Optional[int] = None,
    logprobs: Optional[bool] = None,
    top_logprobs: Optional[int] = None,
    deployment_id=None,
    extra_headers: Optional[dict] = None,
    # soon to be deprecated params by OpenAI
    functions: Optional[List] = None,
    function_call: Optional[str] = None,
    # set api_base, api_version, api_key
    api_version: Optional[str] = None,
    model_list: Optional[list] = None,
    stream: Optional[bool] = None,
    return_messages: Optional[bool] = None,
    verbose: Optional[bool] = None,
) -> List[Response]:
    """Creates a batch completion using the LiteLLM `batch_completion_models_all_responses` endpoint with multiple models."""

    from litellm import batch_completion_models_all_responses

    if not isinstance(model, list):
        raise XNANOException(
            "Batch completions with multiple models requires a list of models."
        )

    if response_model:
        print(
            "[bold]Notice[/bold] - Batch completions default to using LiteLLM `response_format` parameter. Instructor will not be used..."
        )

        response_format = response_model

        if isinstance(response_format, dict):
            response_format = create_dynamic_response_model(response_format)
        elif isinstance(response_format, str):
            response_format = create_dynamic_response_model(response_format)
        elif isinstance(response_format, list):
            response_format = create_dynamic_response_model(response_format)

    new_tools = warn_unsupported_batch_params(tools, run_tools, stream, response_format)

    if new_tools:
        tools = new_tools

    if response_format:
        stream = False

    # format messages
    try:
        messages = format_messages(messages)

        if context:
            messages = add_context_to_messages(messages, context)
    except Exception as e:
        raise XNANOException(
            f"Failed to format messages for multiple model batch completion: {e}"
        )

    if verbose:
        console.message(
            f"✅ [green]Successfully formatted {len(messages)} messages.[/green]"
        )

    response = batch_completion_models_all_responses(
        models=model,
        messages=messages,
        functions=functions,
        function_call=function_call,
        response_format=response_format,
        tools=tools,
        tool_choice=tool_choice,
        parallel_tool_calls=parallel_tool_calls,
        api_key=api_key,
        base_url=base_url,
        organization=organization,
        n=n,
        timeout=timeout,
        temperature=temperature,
        top_p=top_p,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
        logit_bias=logit_bias,
        user=user,
        seed=seed,
        logprobs=logprobs,
        top_logprobs=top_logprobs,
        deployment_id=deployment_id,
        extra_headers=extra_headers,
        max_completion_tokens=max_completion_tokens,
        max_tokens=max_tokens,
        modalities=modalities,
        prediction=prediction,
        audio=audio,
        stream=stream,
    )

    if verbose:
        console.message(
            f"✅ [green]Successfully created batch completion job with {len(messages)} sets of messages and {len(model)} models.[/green]"
        )

    # build structured responses
    if response_format:
        formatted_responses = []

        for response in response:
            formatted_responses.append(
                build_pydantic_model_from_string_dictionary(response, response_format)
            )

        return formatted_responses

    return response



# xnano.resources.agents.agent
# base agent client for xnano

# this script is huge i know sorry
# i will refactor it soon (at some point)

# - imports ----------------------------------------------------
from ...lib import console, XNANOException
from . import helpers
from ..utils import messages as messages_utils
from ..completions.main import completion
from ..completions.resources.tool_calling import convert_to_tool
from .step import Steps

# - types-------------------------------------------------------
from ...types.agents.agent_model import AgentModel
from ...types.agents.agent_completion_args import AgentCompletionArgs
from ...types.agents.state import State
from ...types.agents.agent_response import AgentResponse
from ...types.completions.params import (
    CompletionMessagesParam,
    CompletionChatModelsParam,
    CompletionInstructorModeParam,
    CompletionResponseModelParam,
    CompletionToolChoiceParam,
    CompletionToolsParam,
)
from ...types.completions.responses import Response
from ...types.embeddings.memory import Memory

# - external ----------------------------------------------------
import json
from pydantic import BaseModel, create_model
from typing import Dict, List, Literal, Optional, Union, Type

from litellm import ModelResponse


Agent = Type["Agent"]


# --------------------------------------------------------------
# RESOURCES
# --------------------------------------------------------------


class AgentResources:
    convert_to_tool = convert_to_tool
    messages = messages_utils
    helpers = helpers

    """
    Agent Helper Resources
    """

    # - // messages

    @staticmethod
    def collect_message_from_cli():
        try:
            message = console.input(
                prompt=f"[bold italic] Enter your message:[/bold italic] [bold red]>>[/bold red]  "
            )
        except KeyboardInterrupt:
            return XNANOException(
                message="Keyboard Interrupt.. Try entering a message again"
            )

        return [{"role": "user", "content": message}]

    @staticmethod
    def format_new_messages(messages: CompletionMessagesParam) -> List[Dict]:
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        if isinstance(messages, Dict):
            messages = [messages]

        try:
            messages = AgentResources.messages.verify_messages_integrity(
                messages=messages
            )
        except Exception as e:
            raise XNANOException(message=f"Error formatting new messages: {e}")

        return messages

    # - // responses

    @staticmethod
    def build_response_type(
        response: Response,
        workflow: Optional[BaseModel] = None,
        instructor: bool = False,
    ):
        """
        Types response
        """

        if not instructor:
            return AgentResponse(workflow=workflow, **response.model_dump())
        else:
            return response

    # - // tools

    @staticmethod
    def get_tool_names(
        tools: Optional[CompletionToolsParam] = None,
    ):
        tool_names = []

        if not tools:
            return None

        try:
            for tool in tools:
                if isinstance(tool, str):
                    tool_names.append(tool)
                else:
                    tool = AgentResources.convert_to_tool(tool)
                    tool_names.append(tool.name)
        except Exception as e:
            raise XNANOException(message=f"Error getting tool names: {e}")

        return tool_names

    # - // workflows

    @staticmethod
    def build_empty_workflow_model(workflow: BaseModel) -> BaseModel:
        try:
            workflow_empty_model = create_model(
                workflow.__class__.__name__,
                **{
                    name: (Optional[field.annotation], None)
                    for name, field in workflow.model_fields.items()
                },
            )

            return workflow_empty_model()
        except Exception as e:
            raise XNANOException(message=f"Error building empty workflow model: {e}")

    @staticmethod
    def build_all_workflow_string_descriptions(
        workflows: Optional[List[BaseModel]] = None,
    ) -> str:
        if not workflows:
            return ""

        try:
            workflow_names = []
            workflow_docs = []
            if workflows:
                workflow_names = [workflow.__class__.__name__ for workflow in workflows]
                workflow_docs = [workflow.__class__.__doc__ for workflow in workflows]

        except Exception as e:
            raise XNANOException(f"Error building workflow string descriptions: {e}")

        workflow_string = "\n".join(
            [
                f"{name}: {doc if doc else ''}"
                for name, doc in zip(workflow_names, workflow_docs)
            ]
        )

        return workflow_string


# --------------------------------------------------------------
# AGENT CLASS
# --------------------------------------------------------------


class Agent:
    """
    Base agent class for all agentic workflows/completions & multi
    agent pipelines
    """

    # base level resources
    resources = AgentResources

    def __init__(
        self,
        # agent dsl params
        # required
        role: str = "assistant",
        # optional
        name: Optional[str] = helpers.get_random_name(),
        instructions: Optional[str] = None,
        planning: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
        summarization_steps: Optional[int] = 5,
        agents: Optional[List["Agent"]] = None,
        tools: Optional[List[CompletionToolsParam]] = None,
        temperature: Optional[float] = None,
        # agent memory -- utilized differently than .completion(memory = ...)
        memory: Optional[List[Memory]] = None,
        # agent completion config params
        model: Union[CompletionChatModelsParam, str] = "openai/gpt-4o-mini",
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        messages: Optional[CompletionMessagesParam] = None,
        verbose: bool = False,
    ):
        # VERBOSITY
        # cli outputs for debugging and progress
        self.verbose = verbose

        # STATE
        # handles message and summarization states
        self.state = State(
            messages=messages if messages else [], summary_thread=[], count=0
        )

        # AGENT CONFIG
        self.config = AgentModel(
            role=role,
            name=name,
            instructions=instructions,
            workflows=workflows,
            summarization_steps=summarization_steps,
            planning=planning,
            model=model,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            instructor_mode=instructor_mode,
            agents=agents,
            tools=tools,
            temperature=temperature,
        )

        if self.verbose and agents:
            console.message(
                f"Initialized with [bold gold1]{len(agents)}[/bold gold1] team agents"
            )

        if self.verbose:
            console.message(
                f"New [bold gold1]agent[/bold gold1], [bold red]{name}[/bold red] initialized with role: [bold sky_blue1]{role}[/bold sky_blue1]"
            )

        # ----------------------------------------------------------

    # ----------------------------------------------------------
    # STEPS
    # ----------------------------------------------------------

    def steps(self) -> Steps:
        """Create a new step execution handler"""
        return Steps(self, verbose=self.verbose)

    # ----------------------------------------------------------
    # - internal helper methods
    # ----------------------------------------------------------

    def _get_tools(self, tools: Optional[CompletionToolsParam] = None):
        if tools:
            if self.config.tools:
                return self.config.tools + tools
            else:
                return tools
        else:
            if self.config.tools:
                return self.config.tools
            else:
                return None

    # ----------------------------------------------------------
    # - public helper methods
    # ----------------------------------------------------------

    # - // messages // state

    def add_messages_to_state(self, messages: CompletionMessagesParam):
        try:
            self.state.messages.extend(messages)

            if self.verbose:
                console.message(
                    f"Added [bold gold1]{len(messages)}[/bold gold1] messages to state"
                )
        except Exception as e:
            raise XNANOException(message=f"Error adding messages to state: {e}")

    def get_messages_from_state(self):
        try:
            return self.state.summary_thread + self.state.messages
        except Exception as e:
            raise XNANOException(message=f"Error getting messages from state: {e}")

    def add_response_to_state(
        self,
        response: AgentResponse,
        instructor: bool = False,
    ):
        if instructor:
            self.state.messages.append(
                {"role": "assistant", "content": response.model_dump_json(indent=2)}
            )
            if self.verbose:
                console.message(
                    f"Added [bold sky_blue1]Instructor[/bold sky_blue1] response to state thread for [bold red]{self.config.name}[/bold red]"
                )

            return

        if isinstance(response, BaseModel) and not hasattr(response, "workflow"):
            self.state.messages.append(response.choices[0].message.model_dump())
            return

        if response.workflow is not None:
            self.state.messages.extend(
                [
                    {
                        "role": "assistant",
                        "content": (
                            "I have constructed the following object, now usable for reference"
                            f"\n\n{response.workflow.model_dump_json(indent = 2)}"
                        ),
                    },
                    {
                        "role": "user",
                        "content": ("Use the object to continue your response."),
                    },
                ]
            )
            if self.verbose:
                console.message(
                    f"Added [bold sky_blue1]Workflow[/bold sky_blue1] to state thread for [bold red]{self.config.name}[/bold red]"
                )
        else:
            self.state.messages.append(response.choices[0].message.model_dump())

            if self.verbose:
                console.message(
                    f"Added [bold sky_blue1]Completion[/bold sky_blue1] response to state thread for [bold red]{self.config.name}[/bold red]"
                )

        return

    # - // instruction (system prompt)

    def get_system_prompt(self, tools: Optional[CompletionToolsParam] = None) -> Dict:
        tool_names = AgentResources.get_tool_names(tools)

        instruction_prompt = self.resources.helpers.build_instruction(
            name=self.config.name,
            role=self.config.role,
            tool_names=tool_names,
            instructions=self.config.instructions,
        )

        if self.verbose:
            console.message(
                f"Built instruction system prompt for [bold red]{self.config.name}[/bold red]"
            )

        return instruction_prompt

    # - // summarization

    def build_summary(
        self,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
    ):
        messages = self.state.messages

        try:
            summary = completion(
                messages=[
                    {
                        "role": "system",
                        "content": (
                            "You are a Pulitzer Prize winning journalist. You are tasked with summarizing the following conversation into a concise and informative summary.\n\n"
                            "## INSTRUCTIONS\n"
                            "- Read through the entire given conversation\n"
                            "- Identify the key points, themes and main ideas\n"
                            "- Generate a summary in a detailed list based format\n"
                            "- DO NOT INCLUDE AN INTRODUCTION OR TITLE TO THE SUMMARY\n\n"
                            "## CONVERSATION:"
                        ),
                    },
                    {"role": "user", "content": json.dumps(messages)},
                ],
                model=self.config.model if not model else model,
                api_key=self.config.api_key if not api_key else api_key,
                base_url=self.config.base_url if not base_url else base_url,
                organization=self.config.organization
                if not organization
                else organization,
                verbose=self.verbose,
            )
        except Exception as e:
            raise XNANOException(message=f"Error building summary: {e}")

        self.state.summary_thread.append(
            {
                "role": "user",
                "content": (
                    "The conversation proceeded as follows:\n\n"
                    f"{json.dumps(summary.choices[0].message.content)}"
                ),
            }
        )

        self.state.messages = []
        self.state.count = 0

        return summary.choices[0].message.content

    # - // completion handler

    def _build_completion_request(
        self,
        # required
        messages: CompletionMessagesParam,
        # optional
        # agent dsl params
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        tools: Optional[CompletionToolsParam] = None,
        temperature: Optional[float] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
    ) -> AgentCompletionArgs:
        # add internal state messages to completion messages if applicable
        internal_messages = self.get_messages_from_state()

        if self.verbose:
            console.message(
                f"Found [bold gold1]{len(internal_messages)}[/bold gold1] internal messages"
            )

        if len(internal_messages) > 0:
            messages = internal_messages + messages

        if tools:
            if self.config.tools:
                tools = self.config.tools + tools
        else:
            if self.config.tools:
                tools = self.config.tools

        if not temperature:
            temperature = self.config.temperature

        # build instruction
        instruction_prompt = self.get_system_prompt(tools)

        messages = self.resources.messages.swap_system_prompt(
            messages=messages, system_prompt=instruction_prompt
        )

        args = AgentCompletionArgs(
            messages=messages,
            agents=agents,
            model=model if model else self.config.model,
            base_url=base_url if base_url else self.config.base_url,
            api_key=api_key if api_key else self.config.api_key,
            organization=organization if organization else self.config.organization,
            tools=tools,
            instructor_mode=instructor_mode
            if instructor_mode
            else self.config.instructor_mode,
            response_model=response_model,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            temperature=temperature,
        )

        if self.verbose:
            console.message(
                f"Built completion request for [bold red]{self.config.name}[/bold red]"
            )

        return args

    def _run_completion(
        self,
        args: AgentCompletionArgs,
        response_model: Optional[CompletionResponseModelParam] = None,
    ):
        if self.verbose:
            console.message(
                f"Running completion for [bold red]{self.config.name}[/bold red], with [bold gold1]{len(args.messages)}[/bold gold1] messages"
            )

        try:
            return completion(
                messages=args.messages,
                model=args.model,
                base_url=args.base_url,
                api_key=args.api_key,
                organization=args.organization,
                tools=args.tools,
                mode=args.instructor_mode,
                response_model=response_model
                if response_model
                else args.response_model,
                tool_choice=args.tool_choice,
                parallel_tool_calls=args.parallel_tool_calls,
                context="You have access to the following workflows: \n\n"
                + self.resources.build_all_workflow_string_descriptions(
                    workflows=self.config.workflows
                )
                + "\n Do not hallucinate or make up tools that are not listed.",
                run_tools=True,
            )
        except Exception as e:
            raise XNANOException(message=f"Error running completion: {e}")

    # ----------------------------------------------------------
    # - workflows
    # ----------------------------------------------------------

    def get_workflows(self, workflows: Optional[List[BaseModel]] = None):
        if not workflows:
            return self.config.workflows or []
        if not self.config.workflows:
            return workflows
        return self.config.workflows + workflows

    def _determine_if_workflow_required(
        self, args: AgentCompletionArgs, workflows: Optional[List[BaseModel]] = None
    ) -> bool:
        workflows = self.get_workflows(workflows)

        if len(workflows) == 0:
            return False

        workflow_string = self.resources.build_all_workflow_string_descriptions(
            workflows=workflows
        )

        class WorkflowRequired(BaseModel):
            required: bool

        messages = [
            {
                "role": "system",
                "content": (
                    "You are a world class tool selector. Given a list of tools, you must accurately determine whether or not the current point of the conversation requires the use of a tool.\n\n"
                    f"## TOOLS\n{workflow_string}\n\n"
                ),
            },
            *args.messages,
        ]

        try:
            required = completion(
                messages=messages,
                model=args.model,
                api_key=args.api_key,
                base_url=args.base_url,
                mode=args.instructor_mode,
                response_model=WorkflowRequired,
                verbose=self.verbose,
                temperature=args.temperature,
            )
        except Exception as e:
            raise XNANOException(
                message=f"Error determining if workflow is required: {e}"
            )

        return required.required

    def _select_workflow_to_run(
        self, args: AgentCompletionArgs, workflows: Optional[List[BaseModel]] = None
    ):
        # get workflows
        workflows = self.get_workflows(workflows)

        # build workflow string
        workflow_string = self.resources.build_all_workflow_string_descriptions(
            workflows=workflows
        )

        workflow_names = [
            workflow.model_json_schema()["title"] for workflow in workflows
        ]

        # build workflow selection model
        Selection = create_model(
            "Selection", tool=(Literal[tuple(workflow_names)], ...)
        )

        messages = [
            {
                "role": "system",
                "content": (
                    "You are a world class tool selector. Given a list of tools, you must accurately determine which tool to use.\n\n"
                    f"## TOOLS\n{workflow_string}\n\n"
                ),
            },
            *args.messages,
            {"role": "user", "content": "Do we need to run a tool to continue?"},
            {
                "role": "assistant",
                "content": "I've determined that a tool is required to continue.",
            },
            {
                "role": "user",
                "content": "Please select the most appropriate tool to use.",
            },
        ]

        selection = completion(
            messages=messages,
            model=args.model,
            api_key=args.api_key,
            base_url=args.base_url,
            mode=args.instructor_mode,
            response_model=Selection,
            verbose=self.verbose,
            temperature=args.temperature,
        )

        selected_workflow = selection.tool

        for workflow in workflows:
            if workflow.model_json_schema()["title"] == selected_workflow:
                if self.verbose:
                    console.message(
                        f"Selected workflow: [bold red]{workflow.__class__.__name__}[/bold red]"
                    )

                return workflow

        raise XNANOException(
            f"Selected workflow '{selected_workflow}' not found in workflow list"
        )

    def _execute_workflow(self, workflow: BaseModel, args: AgentCompletionArgs):
        """
        Executes the workflow by generating and executing a plan for each field.

        Args:
            workflow (BaseModel): The workflow model to populate.
            args (AgentCompletionArgs): The arguments for completion.

        Returns:
            BaseModel: The completed workflow object.
        """
        completed_workflow = {}

        if self.verbose:
            console.message(
                f"Executing workflow with planning: [bold red]{workflow.model_json_schema()['title']}[/bold red]"
            )

        # Generate a plan for the workflow execution
        plan = self._plan_workflow_execution(workflow, args)

        # Iterate over each planned step
        messages = args.messages.copy()

        for step in plan.steps:
            field_name = step.field_name
            action_description = step.action

            field_info = workflow.model_fields.get(field_name)
            if not field_info:
                raise XNANOException(
                    message=f"Field '{field_name}' not found in workflow '{workflow.model_json_schema()['title']}'"
                )

            field_type = field_info.annotation

            if self.verbose:
                console.message(
                    f"Executing plan step for field [bold gold1]{field_name}[/bold gold1]: {action_description}"
                )

            # Build messages for this step
            step_messages = messages + [
                {
                    "role": "system",
                    "content": (
                        f"Now working on field '{field_name}' of the workflow '{workflow.model_json_schema()['title']}'. "
                        f"Action: {action_description}"
                    ),
                }
            ]

            # Build response model for the field
            field_response_model = create_model(
                f"{field_name.capitalize()}Response", **{field_name: (field_type, ...)}
            )

            try:
                field_response = completion(
                    messages=step_messages,
                    model=args.model,
                    api_key=args.api_key,
                    base_url=args.base_url,
                    organization=args.organization,
                    instructor_mode=args.instructor_mode,
                    response_model=field_response_model,
                    tools=args.tools if args.tools else self.config.tools,
                    run_tools=True,
                    verbose=self.verbose,
                    temperature=args.temperature,
                )
            except Exception as e:
                raise XNANOException(
                    message=f"Error executing plan step for field '{field_name}': {e}"
                )

            # Add field response to completed workflow
            try:
                completed_workflow[field_name] = getattr(field_response, field_name)

                # Update messages with assistant's response
                messages.append(
                    {
                        "role": "assistant",
                        "content": (
                            f"Completed field '{field_name}':\n"
                            f"{json.dumps(getattr(field_response, field_name), indent=2)}"
                        ),
                    }
                )

            except Exception as e:
                raise XNANOException(
                    message=f"Error adding response for field '{field_name}' to completed workflow: {e}"
                )

        # Instantiate the workflow with completed fields
        try:
            completed_workflow_object = workflow(**completed_workflow)
        except Exception as e:
            raise XNANOException(message=f"Error creating workflow object: {e}")

        if self.verbose:
            console.message(
                f"Completed workflow: [bold sky_blue1]{workflow.model_json_schema()['title']}[/bold sky_blue1] with [bold red]{self.config.name}[/bold red]"
            )

        return completed_workflow_object

    # ----------------------------------------------------------
    # - Planning methods
    # ----------------------------------------------------------

    def _plan_workflow_execution(
        self, workflow: BaseModel, args: AgentCompletionArgs
    ) -> BaseModel:
        """
        Generates a plan for executing each field in the workflow.

        Args:
            workflow (BaseModel): The workflow model to plan.
            args (AgentCompletionArgs): The arguments for completion.

        Returns:
            BaseModel: A plan object containing the steps to execute the workflow.
        """

        # Dynamically create Plan and PlanStep models
        PlanStep = create_model("PlanStep", field_name=(str, ...), action=(str, ...))

        Plan = create_model("Plan", steps=(List[PlanStep], ...))

        # Get workflow schema and description
        workflow_name = workflow.model_json_schema()["title"]
        workflow_description = workflow.model_json_schema().get("description", "")
        workflow_fields = list(workflow.model_fields.keys())

        # Build prompt for planning
        planning_messages = [
            {
                "role": "system",
                "content": (
                    f"You are an expert planner specialized in constructing workflows for {workflow_name}. "
                    f"The workflow is described as: {workflow_description}\n"
                    f"The workflow has the following fields that need to be constructed: {', '.join(workflow_fields)}\n"
                    f"Your task is to generate a detailed plan to build each field of the workflow step by step."
                ),
            },
            {
                "role": "user",
                "content": (
                    "Generate a plan where each step specifies which field to construct and the action required. "
                    "You must include exactly one step for each workflow field, using the exact field names listed above."
                ),
            },
        ]

        try:
            plan = completion(
                messages=planning_messages,
                model=args.model,
                api_key=args.api_key,
                base_url=args.base_url,
                organization=args.organization,
                instructor_mode=args.instructor_mode,
                response_model=Plan,
                verbose=self.verbose,
                temperature=args.temperature,
            )

            # Validate that all field names in the plan match workflow fields
            plan_fields = {step.field_name for step in plan.steps}
            workflow_field_set = set(workflow_fields)

            if plan_fields != workflow_field_set:
                missing = workflow_field_set - plan_fields
                extra = plan_fields - workflow_field_set
                error_msg = []
                if missing:
                    error_msg.append(f"Missing fields: {missing}")
                if extra:
                    error_msg.append(f"Extra fields: {extra}")
                raise XNANOException(
                    message=f"Plan fields don't match workflow fields. {' '.join(error_msg)}"
                )

            if self.verbose:
                console.message(
                    f"Generated workflow execution plan with [bold gold1]{len(plan.steps)}[/bold gold1] steps for [bold red]{workflow_name}[/bold red]."
                )

            return plan

        except Exception as e:
            raise XNANOException(
                message=f"Error generating workflow execution plan: {e}"
            )

    def _generate_plan(
        self,
        messages: CompletionMessagesParam,
    ) -> BaseModel:
        """
        Generates a plan (list of steps) to accomplish the task derived from the messages.

        Args:
            messages (CompletionMessagesParam): The conversation history or user request.

        Returns:
            BaseModel: A plan object containing the steps to accomplish the task.
        """

        # Define the Plan and PlanStep models dynamically
        PlanStep = create_model(
            "PlanStep",
            step_number=(int, ...),
            action=(str, ...),
        )

        Plan = create_model("Plan", steps=(List[PlanStep], ...))

        try:
            # Build prompt to generate the plan
            plan_generation_messages = [
                {
                    "role": "system",
                    "content": (
                        "You are an expert planner. Based on the following conversation, generate a detailed plan "
                        "to accomplish the user's request. The plan should be a numbered list of steps. Each step "
                        "should include a clear and actionable description."
                    ),
                },
                *messages,
            ]

            plan = completion(
                messages=plan_generation_messages,
                model=self.config.model,
                api_key=self.config.api_key,
                base_url=self.config.base_url,
                organization=self.config.organization,
                instructor_mode=self.config.instructor_mode,
                response_model=Plan,
                verbose=self.verbose,
            )

            if self.verbose:
                console.message(
                    f"Generated plan with [bold gold1]{len(plan.steps)}[/bold gold1] steps for [bold red]{self.config.name}[/bold red]."
                )

            return plan

        except Exception as e:
            raise XNANOException(message=f"Error generating plan: {e}")

    def _execute_plan(
        self,
        plan: BaseModel,
        messages: CompletionMessagesParam,
        agents: Optional[List[Agent]] = None,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> List[AgentResponse]:
        """
        Executes the plan by processing each step sequentially.

        Args:
            plan (BaseModel): The plan containing the steps to execute.
            messages (CompletionMessagesParam): The initial messages or conversation context.
            agents (Optional[List[Agent]]): Other agents to collaborate with.
            model (Optional[Union[CompletionChatModelsParam, str]]): Model to use for completions.
            base_url (Optional[str]): Base URL for the API.
            api_key (Optional[str]): API key for authentication.
            organization (Optional[str]): Organization identifier.
            tools (Optional[CompletionToolsParam]): Tools available to the agent.
            instructor_mode (Optional[CompletionInstructorModeParam]): Instructor mode settings.
            response_model (Optional[CompletionResponseModelParam]): Expected response model.
            tool_choice (Optional[CompletionToolChoiceParam]): Tool choice parameters.
            parallel_tool_calls (Optional[bool]): Whether to run tool calls in parallel.
            workflows (Optional[List[BaseModel]]): Workflows available to the agent.

        Returns:
            List[AgentResponse]: A list of responses corresponding to each plan step execution.
        """

        responses = []

        # Copy of the messages to keep the context
        context_messages = messages.copy()

        for step in plan.steps:
            try:
                if self.verbose:
                    console.message(
                        f"Executing plan step [bold gold1]{step.step_number}[/bold gold1]: {step.action}"
                    )

                # Build messages for this step
                step_messages = context_messages + [
                    {
                        "role": "system",
                        "content": (
                            f"Now executing step {step.step_number} of the plan: {step.action}"
                        ),
                    }
                ]

                response = self.run_completion(
                    messages=step_messages,
                    agents=agents,
                    model=model,
                    base_url=base_url,
                    api_key=api_key,
                    organization=organization,
                    tools=tools,
                    instructor_mode=instructor_mode,
                    response_model=response_model,
                    tool_choice=tool_choice,
                    parallel_tool_calls=parallel_tool_calls,
                    workflows=workflows,
                    temperature=temperature,
                )

                responses.append(response)

                # Update context_messages with the assistant's response
                if hasattr(response, "choices") and response.choices:
                    assistant_message = response.choices[0].message.model_dump()
                    context_messages.append(assistant_message)

            except Exception as e:
                raise XNANOException(
                    message=f"Error executing plan step {step.step_number}: {e}"
                )

        if self.verbose:
            console.message(
                f"Executed all [bold gold1]{len(plan.steps)}[/bold gold1] steps for [bold red]{self.config.name}[/bold red]."
            )

        return responses

    def plan(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> List[AgentResponse]:
        """
        Generates a plan to accomplish the given task and executes it step by step.

        Args:
            messages (Optional[CompletionMessagesParam]): Initial messages or user request.
            agents (Optional[List[Agent]]): Other agents to collaborate with.
            model (Optional[Union[CompletionChatModelsParam, str]]): Model to use for completions.
            base_url (Optional[str]): Base URL for the API.
            api_key (Optional[str]): API key for authentication.
            organization (Optional[str]): Organization identifier.
            tools (Optional[CompletionToolsParam]): Tools available to the agent.
            instructor_mode (Optional[CompletionInstructorModeParam]): Instructor mode settings.
            response_model (Optional[CompletionResponseModelParam]): Expected response model.
            tool_choice (Optional[CompletionToolChoiceParam]): Tool choice parameters.
            parallel_tool_calls (Optional[bool]): Whether to run tool calls in parallel.
            workflows (Optional[List[BaseModel]]): Workflows available to the agent.

        Returns:
            List[AgentResponse]: The list of responses from executing each step.
        """

        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        if self.verbose:
            console.message(
                f"Starting planning process for [bold red]{self.config.name}[/bold red]."
            )

        # Generate plan
        plan = self._generate_plan(messages)

        # Execute plan
        responses = self._execute_plan(
            plan=plan,
            messages=messages,
            agents=agents,
            model=model,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            tools=tools,
            instructor_mode=instructor_mode,
            response_model=response_model,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            workflows=workflows,
            temperature=temperature,
        )

        if self.verbose:
            console.message(
                f"Completed planning and execution for [bold red]{self.config.name}[/bold red]."
            )

        return responses

    # ----------------------------------------------------------
    # - multi agent
    # ----------------------------------------------------------

    def _build_agent_query(
        self,
        agent: "Agent",
        messages: CompletionMessagesParam,
    ) -> str:
        """
        Builds a targeted query for a specific agent based on current context
        """
        try:
            query_model = create_model(
                "QueryBuilder", query=(str, ...), reasoning=(str, ...)
            )

            # Build context-aware prompt
            query_messages = [
                {
                    "role": "system",
                    "content": (
                        f"You are an expert query constructor working with {self.config.name}. "
                        f"Your task is to formulate a specific question or request for {agent.config.name}, "
                        f"who has the following role: {agent.config.role}\n\n"
                        "Based on the current conversation context, construct a targeted query that will "
                        "help advance the current objective."
                    ),
                },
                *messages,
                {
                    "role": "user",
                    "content": (
                        "Based on this context, what specific question or request should be asked to "
                        f"{agent.config.name} to help with the current situation? Provide both the query "
                        "and your reasoning."
                        "Ensure that you indicate who you are, and what your role is."
                    ),
                },
            ]

            query_response = completion(
                messages=query_messages,
                model=self.config.model,
                api_key=self.config.api_key,
                base_url=self.config.base_url,
                organization=self.config.organization,
                response_model=query_model,
                mode=self.config.instructor_mode,
                verbose=self.verbose,
            )

            if self.verbose:
                console.message(
                    f"Generated query for [bold red]{agent.config.name}[/bold red] with reasoning: "
                    f"[bold sky_blue1]{query_response.reasoning}[/bold sky_blue1]"
                )

            return query_response.query

        except Exception as e:
            raise XNANOException(message=f"Error building agent query: {e}")

    def _converse_with_agents(
        self,
        agents: List["Agent"],
        messages: CompletionMessagesParam,
    ) -> List[Dict]:
        """
        Manages conversations with multiple agents, collecting their responses
        and formatting them for integration into the main conversation
        """
        if not agents:
            return []

        queries = []
        agent_names = []
        all_responses = []
        workflow_responses = []

        try:
            # Generate queries and get responses from each agent
            for agent in agents:
                query = self._build_agent_query(agent, messages)

                # Ensure query is properly formatted
                if isinstance(query, str):
                    query = [{"role": "user", "content": query}]

                queries.append(query)
                agent_names.append(agent.config.name)

                # Get response from agent using their own configuration
                response = agent.completion(
                    messages=query,
                    instructor_mode=agent.config.instructor_mode,
                    model=agent.config.model,
                    api_key=agent.config.api_key,
                    base_url=agent.config.base_url,
                    organization=agent.config.organization,
                    workflows=agent.config.workflows,
                    temperature=agent.config.temperature,
                )

                # Handle workflow responses separately
                if hasattr(response, "workflow") and response.workflow is not None:
                    workflow_responses.append(
                        f"\nWorkflow from {agent.config.name}:\n"
                        f"```json\n{response.workflow.model_dump_json(indent=2)}\n```"
                    )

                # Format response content
                if hasattr(response, "choices") and response.choices:
                    response_content = response.choices[0].message.content
                else:
                    response_content = response.model_dump_json(indent=2)

                all_responses.append(
                    f"\nResponse from {agent.config.name}:\n{response_content}"
                )

            # Create properly formatted messages
            query_summary = {
                "role": "assistant",
                "content": (
                    "I have consulted with the following agents:\n\n"
                    + "\n".join(
                        [
                            f"- {name}: {query['content'] if isinstance(query, dict) else query}"
                            for name, query in zip(agent_names, queries)
                        ]
                    )
                ),
            }

            consolidated_response = {
                "role": "assistant",
                "content": (
                    "Here are the responses from all consulted agents:"
                    + "".join(all_responses)
                    + (
                        ""
                        if not workflow_responses
                        else "\n\nWorkflows generated:" + "".join(workflow_responses)
                    )
                ),
            }

            # Ensure all messages are properly formatted before returning
            messages_to_return = [query_summary, consolidated_response]
            for msg in messages_to_return:
                if "role" not in msg:
                    msg["role"] = "assistant"
                if "content" not in msg:
                    msg["content"] = str(msg)

            return messages_to_return

        except Exception as e:
            raise XNANOException(message=f"Error in agent conversation: {e}")

    # ----------------------------------------------------------
    # - completions
    # ----------------------------------------------------------

    # - // runner (does not save to state)

    def run_completion(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> AgentResponse:
        completed_workflow = None

        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        # Handle one-time agent collaborations
        if agents:
            agent_messages = self._converse_with_agents(agents, messages)
            messages.extend(agent_messages)

        # Handle team agent consultations
        if self.config.agents:
            required_agents = self._determine_required_agents(messages)
            if required_agents:
                team_messages = self._get_team_agent_responses(
                    messages, required_agents
                )
                messages.extend(team_messages)

        args = self._build_completion_request(
            messages=messages,
            agents=agents,
            model=model,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            tools=tools,
            instructor_mode=instructor_mode,
            response_model=response_model,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            temperature=temperature,
        )

        # Check if workflow is needed
        workflows = self.get_workflows(workflows)
        if workflows and self._determine_if_workflow_required(
            args=args, workflows=workflows
        ):
            # Select and execute workflow
            workflow = self._select_workflow_to_run(args=args, workflows=workflows)
            completed_workflow = self._execute_workflow(workflow=workflow, args=args)

            # add workflow to thread
            self.state.messages.extend(
                [
                    {
                        "role": "assistant",
                        "content": f"{completed_workflow.model_dump_json(indent=2)}",
                    },
                    {
                        "role": "user",
                        "content": f"Use the information to respond to the original user query.",
                    },
                ]
            )

        # Update context to include both workflows and team agents
        team_context = ""
        if self.config.agents:
            team_roles = "\n".join(
                [
                    f"- {agent.config.name}: {agent.config.role}"
                    for agent in self.config.agents
                ]
            )
            team_context = (
                f"\n\nYou have access to the following team members:\n{team_roles}"
            )

        workflow_context = self.resources.build_all_workflow_string_descriptions(
            workflows=self.config.workflows
        )

        context = (
            "You have access to the following workflows: \n\n"
            + workflow_context
            + team_context
            + "\nDo not hallucinate or make up tools/team members that are not listed."
        )

        try:
            response = completion(
                messages=args.messages,
                model=args.model,
                base_url=args.base_url,
                api_key=args.api_key,
                organization=args.organization,
                tools=args.tools,
                instructor_mode=args.instructor_mode,
                response_model=response_model
                if response_model
                else args.response_model,
                tool_choice=args.tool_choice,
                parallel_tool_calls=args.parallel_tool_calls,
                temperature=args.temperature,
                context=context,
                run_tools=True,
                verbose=self.verbose,
            )

            return self.resources.build_response_type(
                response=response,
                workflow=completed_workflow,
                instructor=True if response_model else False,
            )
        except Exception as e:
            raise XNANOException(message=f"Error running completion: {e}")

    # - // main

    def completion(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        workflows: Optional[List[BaseModel]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
    ) -> AgentResponse:
        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        # run completion
        try:
            response = self.run_completion(
                messages=messages,
                agents=agents,
                model=model,
                base_url=base_url,
                api_key=api_key,
                organization=organization,
                tools=tools,
                instructor_mode=instructor_mode,
                response_model=response_model,
                tool_choice=tool_choice,
                parallel_tool_calls=parallel_tool_calls,
                workflows=workflows,
                temperature=temperature,
            )

            # add messages to state on success
            self.add_messages_to_state(messages)

            if response.workflow:
                self.add_messages_to_state(
                    [
                        {
                            "role": "assistant",
                            "content": f"{response.workflow.model_dump_json(indent=2)}",
                        }
                    ]
                )

            # add response to state (handles workflow automatically)
            self.add_response_to_state(
                response=response, instructor=True if response_model else False
            )

        except Exception as e:
            raise XNANOException(message=f"Error running completion: {e}")

        # build summary if needed
        try:
            self.state.count += 1

            if self.state.count >= self.config.summarization_steps:
                self.build_summary(
                    model=model,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                )

        except Exception as e:
            raise XNANOException(message=f"Error incrementing summarization count: {e}")

        return response

    def _determine_required_agents(
        self,
        messages: CompletionMessagesParam,
    ) -> List[str]:
        """
        Determines which agents from the team would be useful for the current context
        """
        if not self.config.agents:
            return []

        try:
            # Ensure messages are properly formatted
            if isinstance(messages, str):
                messages = [{"role": "user", "content": messages}]
            elif isinstance(messages, dict):
                messages = [messages]

            # Create dynamic model for agent selection
            AgentSelection = create_model(
                "AgentSelection", required_agents=(List[str], ...)
            )

            # Build agent roles string
            agent_roles = "\n".join(
                [
                    f"- {agent.config.name}: {agent.config.role}"
                    for agent in self.config.agents
                ]
            )

            # Query for needed agents
            selection = completion(
                messages=[
                    {
                        "role": "system",
                        "content": (
                            "You are an expert team coordinator. Given the available team members "
                            "and the current conversation context, determine which team members would "
                            "be most helpful at this point.\n\n"
                            f"Available team members:\n{agent_roles}\n\n"
                            "Return ONLY the names of team members that would be genuinely useful "
                            "for the current context. Return an empty list if no team members are needed."
                        ),
                    },
                    *messages,
                ],
                model=self.config.model,
                api_key=self.config.api_key,
                base_url=self.config.base_url,
                organization=self.config.organization,
                instructor_mode=self.config.instructor_mode,
                response_model=AgentSelection,
                verbose=self.verbose,
            )

            if self.verbose:
                console.message(
                    f"Selected [bold gold1]{len(selection.required_agents)}[/bold gold1] team agents for current context"
                )

            return selection.required_agents

        except Exception as e:
            raise XNANOException(message=f"Error determining required agents: {e}")

    def _get_team_agent_responses(
        self, messages: CompletionMessagesParam, required_agents: List[str]
    ) -> List[Dict]:
        """
        Gets responses from required team agents
        """
        if not required_agents:
            return []

        selected_agents = [
            agent
            for agent in self.config.agents
            if agent.config.name in required_agents
        ]

        return self._converse_with_agents(selected_agents, messages)

    def workflow(
        self,
        workflow: BaseModel,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        tools: Optional[CompletionToolsParam] = None,
        temperature: Optional[float] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
    ) -> AgentResponse:
        """
        Executes a specific workflow directly without running the workflow selection pipeline.
        Takes the same arguments as completion() but requires a workflow parameter.
        """
        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        # Handle one-time agent collaborations
        if agents:
            agent_messages = self._converse_with_agents(agents, messages)
            messages.extend(agent_messages)

        # Handle team agent consultations
        if self.config.agents:
            required_agents = self._determine_required_agents(messages)
            if required_agents:
                team_messages = self._get_team_agent_responses(
                    messages, required_agents
                )
                messages.extend(team_messages)

        args = self._build_completion_request(
            messages=messages,
            agents=agents,
            model=model,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            tools=tools,
            instructor_mode=instructor_mode,
            response_model=response_model,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            temperature=temperature,
        )

        try:
            completed_workflow = self._execute_workflow(workflow=workflow, args=args)

            response = self.resources.build_response_type(
                response=ModelResponse(
                    choices=[
                        {
                            "message": {
                                "role": "assistant",
                                "content": json.dumps(
                                    completed_workflow.model_dump(), indent=2
                                ),
                            }
                        }
                    ]
                ),
                workflow=completed_workflow,
                instructor=True if response_model else False,
            )

            # Add messages and response to state
            self.add_messages_to_state(messages)
            self.add_response_to_state(
                response=response, instructor=True if response_model else False
            )

            # Handle summarization if needed
            self.state.count += 1
            if self.state.count >= self.config.summarization_steps:
                self.build_summary(
                    model=model,
                    api_key=api_key,
                    base_url=base_url,
                    organization=organization,
                )

            return response

        except Exception as e:
            raise XNANOException(message=f"Error executing workflow: {e}")

    def _analyze_task(
        self,
        messages: CompletionMessagesParam,
    ) -> Dict:
        """Creates simple task definition"""

        class Task(BaseModel):
            task: str
            complete_when: str

        try:
            task = completion(
                messages=[
                    {
                        "role": "system",
                        "content": (
                            "You are a task analyzer. Given the context, define:\n"
                            "1. The core task\n"
                            "2. A single clear condition for completion"
                        ),
                    },
                    *messages,
                ],
                model=self.config.model,
                api_key=self.config.api_key,
                base_url=self.config.base_url,
                organization=self.config.organization,
                instructor_mode=self.config.instructor_mode,
                response_model=Task,
                verbose=self.verbose,
            )

            if self.verbose:
                console.message(
                    f"Defined task: [bold sky_blue1]{task.task}[/bold sky_blue1]"
                )

            return task.model_dump()

        except Exception as e:
            raise XNANOException(message=f"Error analyzing task: {e}")

    def _is_task_complete(
        self,
        messages: List[Dict],
        task: Dict,
    ) -> bool:
        """Simple completion check"""

        class Complete(BaseModel):
            complete: bool

        try:
            check = completion(
                messages=[
                    {
                        "role": "system",
                        "content": (
                            f"Task: {task['task']}\n"
                            f"Complete when: {task['complete_when']}\n\n"
                            "Based on the conversation, is this task complete? "
                            "Return only true or false."
                        ),
                    },
                    *messages,
                ],
                model=self.config.model,
                api_key=self.config.api_key,
                base_url=self.config.base_url,
                organization=self.config.organization,
                instructor_mode=self.config.instructor_mode,
                temperature=self.config.temperature,
                response_model=Complete,
                verbose=self.verbose,
            )

            return check.complete

        except Exception as e:
            raise XNANOException(message=f"Error checking completion: {e}")

    def collaborate(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        max_steps: int = 5,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> AgentResponse:
        """
        Conducts a task-oriented conversation until completion or max steps reached.
        Maintains thread state and integrates with existing agent capabilities.
        """
        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        # Define task
        task = self._analyze_task(messages)

        # Add task context
        messages.append(
            {
                "role": "system",
                "content": (
                    f"Current task: {task['task']}\n"
                    f"Task is complete when: {task['complete_when']}"
                ),
            }
        )

        steps = 0
        last_response = None

        while steps < max_steps:
            try:
                # Check completion after first step
                if steps > 0:
                    state_messages = self.get_messages_from_state()
                    if self._is_task_complete(state_messages, task):
                        break

                # Check if workflow is needed
                workflows = self.get_workflows(workflows)
                if workflows and self._determine_if_workflow_required(
                    args=self._build_completion_request(
                        messages=messages
                        if steps == 0
                        else self.get_messages_from_state(),
                        agents=agents,
                        model=model,
                        base_url=base_url,
                        api_key=api_key,
                        organization=organization,
                        tools=tools,
                        instructor_mode=instructor_mode,
                        response_model=response_model,
                        tool_choice=tool_choice,
                        parallel_tool_calls=parallel_tool_calls,
                        temperature=temperature,
                    ),
                    workflows=workflows,
                ):
                    # Select and execute workflow
                    workflow = self._select_workflow_to_run(
                        args=self._build_completion_request(
                            messages=messages
                            if steps == 0
                            else self.get_messages_from_state(),
                            agents=agents,
                            model=model,
                            base_url=base_url,
                            api_key=api_key,
                            organization=organization,
                            tools=tools,
                            instructor_mode=instructor_mode,
                            response_model=response_model,
                            tool_choice=tool_choice,
                            parallel_tool_calls=parallel_tool_calls,
                            temperature=temperature,
                        ),
                        workflows=workflows,
                    )

                    response = self.workflow(
                        workflow=workflow,
                        messages=messages if steps == 0 else None,
                        agents=agents,
                        model=model,
                        base_url=base_url,
                        api_key=api_key,
                        organization=organization,
                        tools=tools,
                        instructor_mode=instructor_mode,
                        response_model=response_model,
                        tool_choice=tool_choice,
                        parallel_tool_calls=parallel_tool_calls,
                        temperature=temperature,
                    )
                else:
                    # Run regular completion
                    response = self.completion(
                        messages=messages if steps == 0 else None,
                        agents=agents,
                        model=model,
                        base_url=base_url,
                        api_key=api_key,
                        organization=organization,
                        tools=tools,
                        instructor_mode=instructor_mode,
                        response_model=response_model,
                        tool_choice=tool_choice,
                        parallel_tool_calls=parallel_tool_calls,
                        temperature=temperature,
                    )

                last_response = response
                steps += 1
                messages = []  # Clear initial messages after first step

            except Exception as e:
                raise XNANOException(message=f"Error in conversation step {steps}: {e}")

        # Build simple result
        class Result(BaseModel):
            steps: int
            complete: bool
            task: str

        try:
            result = Result(
                steps=steps,
                complete=steps < max_steps,  # True if broke early
                task=task["task"],
            )

            # Format final response
            if response_model:
                return result
            else:
                return self.resources.build_response_type(
                    response=ModelResponse(
                        choices=[
                            {
                                "message": {
                                    "role": "assistant",
                                    "content": json.dumps(
                                        result.model_dump(), indent=2
                                    ),
                                }
                            }
                        ]
                    ),
                    workflow=None,
                )

        except Exception as e:
            raise XNANOException(message=f"Error building conversation result: {e}")

    def update_instructions(self, instructions: str):
        """
        Updates the agent's instructions (system prompt).

        Args:
            instructions (str): New instructions for the agent.
        """
        self.config.instructions = instructions
        if self.verbose:
            console.message(
                f"Updated instructions for [bold red]{self.config.name}[/bold red]"
            )

    def reset_state(self):
        """
        Resets the agent's internal state, including messages and summaries.
        """
        self.state = State(messages=[], summary_thread=[], count=0)
        if self.verbose:
            console.message(f"Reset state for [bold red]{self.config.name}[/bold red]")

    # ----------------------------------------------------------
    # - Planning methods
    # ----------------------------------------------------------

    def _generate_plan(
        self,
        messages: CompletionMessagesParam,
    ) -> BaseModel:
        """
        Generates a plan (list of steps) to accomplish the task derived from the messages.

        Args:
            messages (CompletionMessagesParam): The conversation history or user request.

        Returns:
            BaseModel: A plan object containing the steps to accomplish the task.
        """

        # Define the Plan and PlanStep models dynamically
        PlanStep = create_model(
            "PlanStep",
            step_number=(int, ...),
            action=(str, ...),
        )

        Plan = create_model("Plan", steps=(List[PlanStep], ...))

        try:
            # Build prompt to generate the plan
            plan_generation_messages = [
                {
                    "role": "system",
                    "content": (
                        "You are an expert planner. Based on the following conversation, generate a detailed plan "
                        "to accomplish the user's request. The plan should be a numbered list of steps. Each step "
                        "should include a clear and actionable description."
                    ),
                },
                *messages,
            ]

            plan = completion(
                messages=plan_generation_messages,
                model=self.config.model,
                api_key=self.config.api_key,
                base_url=self.config.base_url,
                organization=self.config.organization,
                instructor_mode=self.config.instructor_mode,
                response_model=Plan,
                verbose=self.verbose,
            )

            if self.verbose:
                console.message(
                    f"Generated plan with {len(plan.steps)} steps for [bold red]{self.config.name}[/bold red]."
                )

            return plan

        except Exception as e:
            raise XNANOException(message=f"Error generating plan: {e}")

    def _execute_plan(
        self,
        plan: BaseModel,
        messages: CompletionMessagesParam,
        agents: Optional[List[Agent]] = None,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> List[AgentResponse]:
        """
        Executes the plan by processing each step sequentially.

        Args:
            plan (BaseModel): The plan containing the steps to execute.
            messages (CompletionMessagesParam): The initial messages or conversation context.
            agents (Optional[List[Agent]]): Other agents to collaborate with.
            model (Optional[Union[CompletionChatModelsParam, str]]): Model to use for completions.
            base_url (Optional[str]): Base URL for the API.
            api_key (Optional[str]): API key for authentication.
            organization (Optional[str]): Organization identifier.
            tools (Optional[CompletionToolsParam]): Tools available to the agent.
            instructor_mode (Optional[CompletionInstructorModeParam]): Instructor mode settings.
            response_model (Optional[CompletionResponseModelParam]): Expected response model.
            tool_choice (Optional[CompletionToolChoiceParam]): Tool choice parameters.
            parallel_tool_calls (Optional[bool]): Whether to run tool calls in parallel.
            workflows (Optional[List[BaseModel]]): Workflows available to the agent.

        Returns:
            List[AgentResponse]: A list of responses corresponding to each plan step execution.
        """

        responses = []

        # Copy of the messages to keep the context
        context_messages = messages.copy()

        for step in plan.steps:
            try:
                if self.verbose:
                    console.message(
                        f"Executing plan step [bold gold1]{step.step_number}[/bold gold1]: {step.action}"
                    )

                # Build messages for this step
                step_messages = context_messages + [
                    {
                        "role": "system",
                        "content": (
                            f"Now executing step {step.step_number} of the plan: {step.action}"
                        ),
                    }
                ]

                response = self.run_completion(
                    messages=step_messages,
                    agents=agents,
                    model=model,
                    base_url=base_url,
                    api_key=api_key,
                    organization=organization,
                    tools=tools,
                    instructor_mode=instructor_mode,
                    response_model=response_model,
                    tool_choice=tool_choice,
                    parallel_tool_calls=parallel_tool_calls,
                    workflows=workflows,
                    temperature=temperature,
                )

                responses.append(response)

                # Update context_messages with the assistant's response
                if hasattr(response, "choices") and response.choices:
                    assistant_message = response.choices[0].message.model_dump()
                    context_messages.append(assistant_message)

            except Exception as e:
                raise XNANOException(
                    message=f"Error executing plan step {step.step_number}: {e}"
                )

        if self.verbose:
            console.message(
                f"Executed all [bold gold1]{len(plan.steps)}[/bold gold1] steps for [bold red]{self.config.name}[/bold red]."
            )

        return responses

    def plan(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> List[AgentResponse]:
        """
        Generates a plan to accomplish the given task and executes it step by step.

        Args:
            messages (Optional[CompletionMessagesParam]): Initial messages or user request.
            agents (Optional[List[Agent]]): Other agents to collaborate with.
            model (Optional[Union[CompletionChatModelsParam, str]]): Model to use for completions.
            base_url (Optional[str]): Base URL for the API.
            api_key (Optional[str]): API key for authentication.
            organization (Optional[str]): Organization identifier.
            tools (Optional[CompletionToolsParam]): Tools available to the agent.
            instructor_mode (Optional[CompletionInstructorModeParam]): Instructor mode settings.
            response_model (Optional[CompletionResponseModelParam]): Expected response model.
            tool_choice (Optional[CompletionToolChoiceParam]): Tool choice parameters.
            parallel_tool_calls (Optional[bool]): Whether to run tool calls in parallel.
            workflows (Optional[List[BaseModel]]): Workflows available to the agent.

        Returns:
            List[AgentResponse]: The list of responses from executing each step.
        """

        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        if self.verbose:
            console.message(
                f"Starting planning process for [bold red]{self.config.name}[/bold red]."
            )

        # Generate plan
        plan = self._generate_plan(messages)

        # Execute plan
        responses = self._execute_plan(
            plan=plan,
            messages=messages,
            agents=agents,
            model=model,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            tools=tools,
            instructor_mode=instructor_mode,
            response_model=response_model,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            workflows=workflows,
            temperature=temperature,
        )

        if self.verbose:
            console.message(
                f"Completed planning and execution for [bold red]{self.config.name}[/bold red]."
            )

        return responses

    def chat_completion(
        self,
        messages: Optional[CompletionMessagesParam] = None,
        agents: Optional[List[Agent]] = None,
        # completion specific params
        model: Optional[Union[CompletionChatModelsParam, str]] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        organization: Optional[str] = None,
        temperature: Optional[float] = None,
        tools: Optional[CompletionToolsParam] = None,
        instructor_mode: Optional[CompletionInstructorModeParam] = None,
        response_model: Optional[CompletionResponseModelParam] = None,
        tool_choice: Optional[CompletionToolChoiceParam] = None,
        parallel_tool_calls: Optional[bool] = False,
        workflows: Optional[List[BaseModel]] = None,
    ) -> AgentResponse:
        if not messages:
            messages = self.resources.collect_message_from_cli()

        messages = self.resources.format_new_messages(messages)

        # Handle one-time agent collaborations
        if agents:
            agent_messages = self._converse_with_agents(agents, messages)
            messages.extend(agent_messages)

        # Handle team agent consultations
        if self.config.agents:
            required_agents = self._determine_required_agents(messages)
            if required_agents:
                team_messages = self._get_team_agent_responses(
                    messages, required_agents
                )
                messages.extend(team_messages)

        args = self._build_completion_request(
            messages=messages,
            agents=agents,
            model=model,
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            tools=tools,
            instructor_mode=instructor_mode,
            response_model=response_model,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            temperature=temperature,
        )

        # Update context to include both workflows and team agents
        team_context = ""
        if self.config.agents:
            team_roles = "\n".join(
                [
                    f"- {agent.config.name}: {agent.config.role}"
                    for agent in self.config.agents
                ]
            )
            team_context = (
                f"\n\nYou have access to the following team members:\n{team_roles}"
            )

        workflow_context = self.resources.build_all_workflow_string_descriptions(
            workflows=self.config.workflows
        )

        context = (
            "You have access to the following workflows: \n\n"
            + workflow_context
            + team_context
            + "\nDo not hallucinate or make up tools/team members that are not listed."
        )

        try:
            response = completion(
                messages=args.messages,
                model=args.model if args.model else self.config.model,
                base_url=args.base_url if args.base_url else self.config.base_url,
                api_key=args.api_key if args.api_key else self.config.api_key,
                organization=args.organization
                if args.organization
                else self.config.organization,
                tools=args.tools if args.tools else self.config.tools,
                instructor_mode=args.instructor_mode
                if args.instructor_mode
                else self.config.instructor_mode,
                response_model=response_model
                if response_model
                else args.response_model,
                tool_choice=args.tool_choice,
                parallel_tool_calls=args.parallel_tool_calls,
                context=context,
                run_tools=True,
                verbose=self.verbose,
                temperature=temperature if temperature else self.config.temperature,
            )

            return self.resources.build_response_type(
                response=response,
                workflow=None,
                instructor=True if response_model else False,
            )
        except Exception as e:
            raise XNANOException(message=f"Error running chat completion: {e}")



from pydantic import BaseModel
from typing import List, Optional


class AgentModel(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    # agent definition params
    name: Optional[str] = None
    role: str
    instructions: Optional[str] = None

    agents: Optional[List] = None

    # agent specialization params
    tools: Optional[List] = None
    workflows: Optional[List] = None
    memory: Optional[List] = None

    # agent llm params
    model: str = "gpt-4o-mini"
    base_url: Optional[str] = None
    api_key: Optional[str] = None
    temperature: Optional[float] = None
    organization: Optional[str] = None
    instructor_mode: Optional[str] = None

    # agent summary params
    summarization_steps: Optional[int] = 5

    # workflow / planning params
    planning: bool = False



__all__ = [
    "generate_code",
    "generate_function",
    "generate_classification",
    "async_generate_classification",
    "generate_chunks",
    "generate_extraction",
    "async_generate_extraction",
    "generate_sql",
    "generate_system_prompt",
    "generate_qa_pairs",
    "generate_answers",
    "generate_questions",
    "generate_validation",
    "async_generate_validation",
    # multimodal
    "multimodal_generate_image",
    "multimodal_generate_audio",
    "multimodal_generate_transcription",
]

from .code_generators import generate_code, generate_function
from .classifier import generate_classification, async_generate_classification
from .chunker import generate_chunks
from .extractor import generate_extraction, async_generate_extraction
from .generate_sql import generate_sql
from .prompting import generate_system_prompt
from .question_answer import generate_qa_pairs, generate_answers, generate_questions
from .validator import generate_validation, async_generate_validation

# multimodal
from .multimodal import (
    multimodal_generate_image,
    multimodal_generate_audio,
    multimodal_generate_transcription,
)



from pydantic import BaseModel, Field, validator
from typing import Dict, List, Optional, Any, Callable, Union, TypeVar
from enum import Enum

T = TypeVar("T")


class StepStatus(str, Enum):
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    SKIPPED = "skipped"


class StepModel(BaseModel):
    """Base model for step configuration"""

    name: str
    handler: Callable
    depends_on: Optional[List[str]] = None
    condition: Optional[Callable] = None
    status: StepStatus = StepStatus.PENDING
    result: Optional[Any] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)

    @validator("depends_on")
    def validate_dependencies(cls, v):
        if v is not None and not isinstance(v, list):
            raise ValueError("depends_on must be a list of step names")
        return v


class StepState(BaseModel):
    """State management for steps"""

    messages: List[Dict[str, Any]] = Field(default_factory=list)
    results: Dict[str, Any] = Field(default_factory=dict)
    metadata: Dict[str, Any] = Field(default_factory=dict)

    def add_result(self, step_name: str, result: Any):
        """Safely add a result to the state"""
        self.results[step_name] = result

    def get_result(self, step_name: str) -> Any:
        """Safely retrieve a result from the state"""
        return self.results.get(step_name)



# Agent Completion Args

from pydantic import BaseModel
from typing import Optional, Dict, List, Union, Any


class AgentCompletionArgs(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    messages: Optional[Union[List, Dict, str]] = None

    model: Optional[str] = None

    api_key: Optional[str] = None

    organization: Optional[str] = None

    base_url: Optional[str] = None

    temperature: Optional[float] = None

    agents: Optional[Union[List[Any], Any]] = None

    tools: Optional[List[Any]] = None

    instructor_mode: Optional[str] = None

    response_model: Optional[Any] = None

    tool_choice: Optional[str] = None

    parallel_tool_calls: Optional[bool] = None



__all__ = ["text_reader", "text_embeddings", "text_chunker"]

from .documents.text_reader import text_reader
from .embeddings.text_embeddings import text_embeddings
from .processing.text_chunker import text_chunker



import json
from pathlib import Path
from typing import Dict, List, Tuple, Optional
from .console import console
import shutil
from datetime import datetime


class SystemPromptCache:
    def __init__(self, path: Path):
        self.path = path
        self.path.mkdir(parents=True, exist_ok=True)

    def get_prompt_ids(self) -> List[str]:
        """Get all available prompt IDs"""
        console.message(f"Getting prompt ids from {self.path}")
        return [f.stem for f in self.path.iterdir() if f.is_file()]

    def get_prompt(self, prompt_id: str) -> Dict[str, str]:
        """Get a system prompt by its ID"""
        prompt_file = self.path / f"{prompt_id}.json"
        if not prompt_file.exists():
            raise KeyError(f"No prompt found with ID: {prompt_id}")

        with open(prompt_file, "r") as f:
            return json.load(f)

    def save_prompt(
        self, prompt_id: str, prompt: str, name: str, description: Optional[str] = None
    ):
        """Save or update a system prompt with the given ID and metadata"""
        prompt_data = {
            "name": name,
            "description": description,
            "prompt": prompt,
            "created_at": datetime.now().isoformat(),
        }
        with open(self.path / f"{prompt_id}.json", "w") as f:
            json.dump(prompt_data, f)


class MessageCache:
    def __init__(self, messages_dir: Path):
        self.messages_dir = messages_dir
        self.messages_dir.mkdir(parents=True, exist_ok=True)

    def get_messages(self, thread_id: str) -> List[Dict]:
        """Get messages for a specific thread"""
        console.message(f"Getting messages for thread {thread_id}")
        message_file = self.messages_dir / f"{thread_id}.json"

        if not message_file.exists():
            return []

        with open(message_file, "r") as f:
            thread_data = json.load(f)
            return thread_data["messages"]

    def add_message(self, message: Dict, thread_id: str):
        """Add a message to a specific thread"""
        console.message(f"Adding message to thread {thread_id}")
        message_file = self.messages_dir / f"{thread_id}.json"

        if message_file.exists():
            with open(message_file, "r") as f:
                thread_data = json.load(f)
                messages = thread_data["messages"]
        else:
            thread_data = {"created_at": datetime.now().isoformat(), "messages": []}
            messages = thread_data["messages"]

        messages.append(message)
        thread_data["updated_at"] = datetime.now().isoformat()

        with open(message_file, "w") as f:
            json.dump(thread_data, f)

    def get_threads(self) -> List[Dict]:
        """
        Get all threads from the cache
        Returns a list of thread dictionaries with messages and metadata
        """
        threads = []
        for thread_file in self.messages_dir.glob("*.json"):
            try:
                with open(thread_file, "r") as f:
                    thread_data = json.load(f)
                    threads.append(
                        {
                            "thread_id": thread_file.stem,
                            "created_at": thread_data.get("created_at", ""),
                            "updated_at": thread_data.get("updated_at", ""),
                            "messages": thread_data.get("messages", []),
                        }
                    )
            except json.JSONDecodeError:
                console.error(f"Error reading thread file: {thread_file}")
                continue

        # Sort threads by updated_at date, newest first
        return sorted(threads, key=lambda x: x.get("updated_at", ""), reverse=True)


class Cache:
    def __init__(self):
        console.message("Initializing cache")
        self.cache_dir = Path.home() / ".cache" / "xnano"
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        # Create cache subdirectories
        self.messages_dir = self.cache_dir / "messages"
        self.system_prompts_dir = self.cache_dir / "system_prompts"

        self.messages_dir.mkdir(parents=True, exist_ok=True)
        self.system_prompts_dir.mkdir(parents=True, exist_ok=True)

        # Initialize caches
        self.system_prompt_cache = SystemPromptCache(self.system_prompts_dir)
        self.message_cache = MessageCache(self.messages_dir)

    def get_thread_ids(self) -> List[str]:
        """Get all thread IDs"""
        console.message("Getting thread ids")
        return self.message_cache.get_threads()

    def clear_message_cache(self, thread_id: str):
        """Clear message cache for a specific thread"""
        console.message(f"Clearing message cache for {thread_id}")
        message_file = self.messages_dir / f"{thread_id}.json"
        message_file.unlink(missing_ok=True)

    def reset(self):
        """Reset the entire cache"""
        if self.cache_dir.exists():
            shutil.rmtree(self.cache_dir)
        self.__init__()


cache = Cache()



from typing import Literal, Optional, Any, Union
from ...lib import console


__all__ = ["image", "audio", "transcribe"]


def prompt_cli_install(library: str):
    """Installs a specified library"""

    import os

    os.system(f"pip install {library}")


ModelType = Literal[
    "dall-e-2",
    "dall-e-3",
    "flux-dev",
    "flux-realism",
    "flux-schnell",
    "flux-pro",
    "flux-lora",
    "flux-general",
    "aura",
    "sd-v3",
    "fooocus",
]

PREDEFINED_APPS: Literal[
    "fal-ai/flux/dev",
    "fal-ai/flux-realism",
    "fal-ai/flux/schnell",
    "fal-ai/flux-pro",
    "fal-ai/flux-lora",
    "fal-ai/flux-general",
    "fal-ai/aura-flow",
    "fal-al/lora",
    "fal-ai/stable-diffusion-v3-medium",
    "fal-ai/fooocus",
] = None


def _get_model_config(model: ModelType):
    dall_e_models = ["dall-e-2", "dall-e-3"]
    fal_models = {
        "flux-dev": "fal-ai/flux/dev",
        "flux-realism": "fal-ai/flux-realism",
        "flux-schnell": "fal-ai/flux/schnell",
        "flux-pro": "fal-ai/flux-pro",
        "flux-lora": "fal-ai/flux-lora",
        "flux-general": "fal-ai/flux-general",
        "aura": "fal-ai/aura-flow",
        "sd-v3": "fal-ai/stable-diffusion-v3-medium",
        "fooocus": "fal-ai/fooocus",
    }

    if model in dall_e_models:
        return {"provider": "openai", "model": model}
    elif model in fal_models:
        return {"provider": "fal", "application": fal_models[model]}
    else:
        raise ValueError(f"Unsupported model: {model}")


def multimodal_generate_image(
    prompt: str,
    model: ModelType = "dall-e-3",
    api_key: Optional[str] = None,
    image_size: Optional[str] = "landscape_4_3",
    num_inference_steps: Optional[int] = 26,
    guidance_scale: Optional[float] = 3.5,
    enable_safety_checker: Optional[bool] = False,
    size: Optional[str] = "1024x1024",
    quality: Optional[str] = "standard",
    n: Optional[int] = 1,
    display: Optional[bool] = False,
    optimize_prompt: Optional[bool] = False,
    optimize_prompt_model: Optional[str] = "openai/gpt-4o-mini",
) -> Union[str, Any]:
    """Generates an image using either the FAL_AI API or OpenAI. With an
    optional display function to show the image in a notebook.

    Parameters:
        prompt: str,
        model: ModelType = "dall-e-3",
        api_key: Optional[str] = None,
        image_size: Optional[str] = "landscape_4_3",
        num_inference_steps: Optional[int] = 26,
        guidance_scale: Optional[float] = 3.5,
        enable_safety_checker: Optional[bool] = False,
        size: Optional[str] = "1024x1024",
        quality: Optional[str] = "standard",
        n: Optional[int] = 1,
        display: Optional[bool] = False,

    Returns:
        str or Any: The generated image or an error message.
    """
    model_config = _get_model_config(model)

    if model_config["provider"] == "openai":
        from openai import OpenAI

        try:
            client = OpenAI(api_key=api_key)
        except Exception as e:
            return e
        try:
            response = client.images.generate(
                model=model_config["model"],
                prompt=prompt,
                size=size,
                quality=quality,
                n=n,
            )
        except Exception as e:
            return e
        if display:
            try:
                from IPython.display import display, Image
            except ImportError:
                console.warning(
                    "IPython is not installed. Please install it with `pip install ipython`."
                )
                prompt_cli_install("IPython")

            url = response.data[0].url
            display(Image(url=url))
        return response

    elif model_config["provider"] == "fal":
        try:
            import fal_client
        except ImportError:
            console.critical(
                "FAL_AI API requires the 'fal-client' package. Please install it with `pip install fal-client`."
            )
            prompt_cli_install("fal-client")

        if optimize_prompt:
            from ..completions.main import completion
            from pydantic import BaseModel

            class OptimizedPrompt(BaseModel):
                prompt: str

            optimized_prompt = completion(
                messages=[
                    {
                        "role": "system",
                        "content": f"""
                 ## CONTEXT ## \n
                 You are a world class image description optimizer. You enhance descriptions of images at incredible quality, detail, but with a focus on being concise. You define your descritpions
                 in a comma list of 2-3 word phrases. \n\n
                 
                 ## INSTRUCTIONS ## \n
                 - You will be given a description of an image
                 - Reason about the image as a whole, and descriptions the user has provided
                 - Optimize the prompt for use in image generation
                 - Ensure that the optimized prompt is a concise, detailed list of 2-3 word phrases.
                 
                 ## EXAMPLE ## \n
                 Original Prompt : [ A beautiful landscape painting of a sunset over the ocean. ] \n
                 Optimized Prompt : [ A beautiful painting, pink vibrant sunset, dynamic ocean waves, vibrant art, 4k, brush strokes ]
                 
                 """,
                    },
                    {
                        "role": "user",
                        "content": f"Optimize this image description for use in image generation. The original prompt is : [ {prompt} ]",
                    },
                ],
                model=optimize_prompt_model,
                response_model=OptimizedPrompt,
            )

            prompt = optimized_prompt.prompt

        try:
            handler = fal_client.submit(
                application=model_config["application"],
                arguments={
                    "prompt": prompt,
                    "image_size": image_size,
                    "num_inference_steps": num_inference_steps,
                    "guidance_scale": guidance_scale,
                    "enable_safety_checker": enable_safety_checker,
                    "num_images": n,
                },
            )
            result = handler.get()
            if display:
                try:
                    from IPython.display import display, Image
                except ImportError:
                    console.warning(
                        "IPython is not installed. Please install it with `pip install ipython`."
                    )
                    prompt_cli_install("IPython")

                url = result["images"][0]["url"]
                display(Image(url=url))
        except Exception as e:
            result = e
        return result


OPENAI_TTS_MODELS = Literal["tts-1", "tts-1-hd"]
OPENAI_TTS_VOICES = Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"]


def multimodal_generate_audio(
    prompt: str,
    model: OPENAI_TTS_MODELS = "tts-1",
    voice: OPENAI_TTS_VOICES = "alloy",
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    filename: Optional[str] = None,
    play: bool = False,
):
    """Generates an audio file from text, through the openai API.

    Parameters:
        prompt: str,
        model: OPENAI_TTS_MODELS = "tts-1",
        voice: OPENAI_TTS_VOICES = "alloy",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        filename: Optional[str] = None,
        play: bool = False,

    Returns:
        str or Any: The generated audio file or an error message.
    """
    from openai import OpenAI
    import io

    try:
        import sounddevice as sd
        import soundfile as sf

    except ImportError:
        console.warning(
            "The [italic]speak[/italic] function requires sounddevice and soundfile, which are not included in the base 'xnano' package. Please install them with [bold]`pip install sounddevice soundfile`[/bold]."
        )
        prompt_cli_install("sounddevice soundfile")

    client = OpenAI(api_key=api_key, base_url=base_url)
    try:
        response = client.audio.speech.create(input=prompt, model=model, voice=voice)
        audio_data = response.read()

        try:
            with io.BytesIO(audio_data) as audio_buffer:
                audio_array, sample_rate = sf.read(audio_buffer)
        except Exception as e:
            return e

        if filename:
            file_endings = [".wav", ".mp3", ".m4a"]
            if not filename.endswith(tuple(file_endings)):
                raise ValueError(
                    f"Filename must end with one of the following: {', '.join(file_endings)}"
                )

            sf.write(filename, audio_array, sample_rate)

        if play:
            try:
                from IPython.display import Audio
            except ImportError:
                console.warning(
                    "IPython is not installed. Please install it with `pip install ipython`."
                )
                prompt_cli_install("IPython")
            # Play audio using sounddevice
            sd.play(audio_array, sample_rate)
            sd.wait()

            # For Jupyter notebook, also return IPython audio widget
            return Audio(audio_array, rate=sample_rate, autoplay=True)
        else:
            return audio_array, sample_rate

    except Exception as e:
        return str(e)


def multimodal_generate_transcription(
    model: str = "whisper-1",
    api_key: Optional[str] = None,
    base_url: Optional[str] = None,
    organization: Optional[str] = None,
    file: Optional[str] = None,
    record: bool = False,
    duration: int = 5,
):
    """Transcribes an audio file into text, through the openai API.

    Parameters:
        model: str = "whisper-1",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        file: Optional[str] = None,
        record: bool = False,
        duration: int = 5,

    Returns:
        str or Any: The transcribed text or an error message.
    """
    from openai import OpenAI

    try:
        import sounddevice as sd
        import soundfile as sf
    except ImportError:
        console.warning(
            "The [italic]speak[/italic] function requires sounddevice and soundfile, which are not included in the base 'xnano` package. Please install them with [bold]`pip install sounddevice soundfile`[/bold]."
        )
        prompt_cli_install("sounddevice soundfile")

    import io

    client = OpenAI(api_key=api_key, base_url=base_url, organization=organization)

    if record:
        print(f"Recording for {duration} seconds...")
        audio_data = sd.rec(int(duration * 44100), samplerate=44100, channels=1)
        sd.wait()
        print("Recording finished.")

        with io.BytesIO() as buffer:
            sf.write(buffer, audio_data, 44100, format="wav")
            buffer.seek(0)
            audio_file = buffer
    elif file:
        if not file.endswith((".mp3", ".wav", ".m4a")):
            raise ValueError("File must be a .mp3, .wav, or .m4a file")
        audio_file = open(file, "rb")
    else:
        raise ValueError(
            "Either 'file' must be provided or 'record' must be set to True"
        )

    try:
        transcription = client.audio.transcriptions.create(
            model=model, file=audio_file, response_format="text"
        )
        return transcription
    except Exception as e:
        return str(e)


class MultimodalClient:
    @staticmethod
    def multimodal_generate_image(
        prompt: str,
        model: ModelType = "dall-e-3",
        api_key: Optional[str] = None,
        image_size: Optional[str] = "landscape_4_3",
        num_inference_steps: Optional[int] = 26,
        guidance_scale: Optional[float] = 3.5,
        enable_safety_checker: Optional[bool] = False,
        size: Optional[str] = "1024x1024",
        quality: Optional[str] = "standard",
        n: Optional[int] = 1,
        display: Optional[bool] = False,
        optimize_prompt: Optional[bool] = False,
        optimize_prompt_model: Optional[str] = "openai/gpt-4o-mini",
    ):
        return multimodal_generate_image(
            prompt=prompt,
            model=model,
            api_key=api_key,
            image_size=image_size,
            num_inference_steps=num_inference_steps,
            guidance_scale=guidance_scale,
            enable_safety_checker=enable_safety_checker,
            size=size,
            quality=quality,
            n=n,
            display=display,
            optimize_prompt=optimize_prompt,
            optimize_prompt_model=optimize_prompt_model,
        )

    @staticmethod
    def multimodal_generate_audio(
        prompt: str,
        model: OPENAI_TTS_MODELS = "tts-1",
        voice: OPENAI_TTS_VOICES = "alloy",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        filename: Optional[str] = None,
        play: bool = False,
    ):
        return multimodal_generate_audio(
            prompt=prompt,
            model=model,
            voice=voice,
            api_key=api_key,
            base_url=base_url,
            filename=filename,
            play=play,
        )

    @staticmethod
    def multimodal_generate_transcription(
        model: str = "whisper-1",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        organization: Optional[str] = None,
        file: Optional[str] = None,
        record: bool = False,
        duration: int = 5,
    ):
        return multimodal_generate_transcription(
            model=model,
            api_key=api_key,
            base_url=base_url,
            organization=organization,
            file=file,
            record=record,
            duration=duration,
        )


if __name__ == "__main__":
    print(
        multimodal_generate_image(
            model="flux-dev",
            prompt="A beautiful landscape painting of a sunset over the ocean.",
            n=1,
        )
    )



from ...types.agents.step import StepModel, StepState, StepStatus
from ...lib import console, XNANOException
from typing import Dict, List, Optional, Any, Callable, Type, Union
from functools import wraps

Agent = Type["Agent"]


class Steps:
    """
    Step execution handler for agents
    """

    def __init__(self, agent: Optional[Agent] = None, verbose: bool = False):
        self.agent = agent
        self.verbose = verbose
        self.steps: Dict[str, StepModel] = {}
        self.state = StepState()
        self._current_step: Optional[str] = None
        self._executed_steps: List[str] = []

    def step(
        self,
        name: str,
        depends_on: Optional[List[str]] = None,
        condition: Optional[Callable[[StepState], bool]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Callable:
        """
        Step decorator that can be used both with and without an agent

        Example:
            @steps.step("process_data", depends_on=["fetch_data"])
            def process_data(state: StepState, input_data: Dict[str, Any]) -> Any:
                # Process the data
                return processed_data
        """

        def decorator(func: Callable) -> Callable:
            @wraps(func)
            def wrapper(*args, **kwargs):
                # Handle both agent and standalone contexts
                if self.agent:
                    return func(
                        agent=self.agent,
                        state=self.state,
                        input_data=kwargs.get("input_data", {}),
                    )
                return func(state=self.state, input_data=kwargs.get("input_data", {}))

            self.steps[name] = StepModel(
                name=name,
                handler=wrapper,
                depends_on=depends_on,
                condition=condition,
                metadata=metadata or {},
            )

            if self.verbose:
                agent_name = self.agent.config.name if self.agent else "standalone"
                console.message(
                    f"Added step [bold gold1]{name}[/bold gold1] to [bold red]{agent_name}[/bold red]"
                )

            return wrapper

        return decorator

    def add_step(
        self,
        name: str,
        handler: Callable,
        depends_on: Optional[List[str]] = None,
        condition: Optional[Callable[[StepState], bool]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Programmatically add a step without using the decorator

        Example:
            steps.add_step("process_data", process_data_func, depends_on=["fetch_data"])
        """
        if name in self.steps:
            raise XNANOException(f"Step {name} already exists")

        @wraps(handler)
        def wrapped_handler(*args, **kwargs):
            if self.agent:
                return handler(
                    agent=self.agent,
                    state=self.state,
                    input_data=kwargs.get("input_data", {}),
                )
            return handler(state=self.state, input_data=kwargs.get("input_data", {}))

        self.steps[name] = StepModel(
            name=name,
            handler=wrapped_handler,
            depends_on=depends_on,
            condition=condition,
            metadata=metadata or {},
        )

    def _validate_dependencies(self) -> None:
        """Validate all step dependencies exist and there are no cycles"""
        for step in self.steps.values():
            if step.depends_on:
                for dep in step.depends_on:
                    if dep not in self.steps:
                        raise XNANOException(
                            f"Step {step.name} depends on non-existent step {dep}"
                        )

        # Check for cycles
        visited = set()
        temp = set()

        def has_cycle(step_name: str) -> bool:
            if step_name in temp:
                return True
            if step_name in visited:
                return False

            temp.add(step_name)
            step = self.steps[step_name]
            if step.depends_on:
                for dep in step.depends_on:
                    if has_cycle(dep):
                        return True
            temp.remove(step_name)
            visited.add(step_name)
            return False

        for step_name in self.steps:
            if has_cycle(step_name):
                raise XNANOException("Circular dependency detected in steps")

    def _can_run_step(self, step: StepModel) -> bool:
        """Check if a step can be executed"""
        if not step.depends_on:
            return True
        return all(
            self.steps[dep].status == StepStatus.COMPLETED for dep in step.depends_on
        )

    def _execute_step(self, step: StepModel) -> Any:
        """Execute a single step with proper error handling and state management"""
        try:
            if step.name in self._executed_steps:
                return step.result

            step.status = StepStatus.RUNNING
            self._current_step = step.name

            # Check condition
            if step.condition and not step.condition(self.state):
                step.status = StepStatus.SKIPPED
                return None

            # Get input data from dependencies
            input_data = {
                dep: self.steps[dep].result
                for dep in (step.depends_on or [])
                if self.steps[dep].status == StepStatus.COMPLETED
            }

            # Execute handler
            result = step.handler(input_data=input_data)

            step.result = result
            step.status = StepStatus.COMPLETED
            self.state.add_result(step.name, result)
            self._executed_steps.append(step.name)

            if self.verbose:
                agent_name = self.agent.config.name if self.agent else "standalone"
                console.message(
                    f"Completed step [bold gold1]{step.name}[/bold gold1] for [bold red]{agent_name}[/bold red]"
                )

            return result

        except Exception as e:
            step.status = StepStatus.FAILED
            raise XNANOException(message=f"Error executing step {step.name}: {str(e)}")
        finally:
            self._current_step = None

    def execute(self) -> Dict[str, Any]:
        """
        Execute all steps in the correct order

        Returns:
            Dict[str, Any]: Results from all completed steps
        """
        if not self.steps:
            raise XNANOException(message="No steps defined")

        self._validate_dependencies()
        self._executed_steps = []

        while True:
            # Get all runnable steps for potential parallel execution
            runnable_steps = [
                step
                for step in self.steps.values()
                if step.status == StepStatus.PENDING and self._can_run_step(step)
            ]

            if not runnable_steps:
                if all(
                    s.status == StepStatus.COMPLETED or s.status == StepStatus.SKIPPED
                    for s in self.steps.values()
                ):
                    break
                if any(s.status == StepStatus.FAILED for s in self.steps.values()):
                    failed_steps = [
                        s.name
                        for s in self.steps.values()
                        if s.status == StepStatus.FAILED
                    ]
                    raise XNANOException(
                        message=f"Step execution failed. Failed steps: {', '.join(failed_steps)}"
                    )
                raise XNANOException(
                    message="Step execution deadlocked - possible circular dependency"
                )

            # Execute each runnable step
            for step in runnable_steps:
                self._execute_step(step)

        return self.state.results

    def reset(self) -> None:
        """Reset all steps to their initial state"""
        for step in self.steps.values():
            step.status = StepStatus.PENDING
            step.result = None
        self.state = StepState()
        self._executed_steps = []
        self._current_step = None



# Completion Tests

import asyncio
from typing import Generator
from xnano import completion, async_completion
from pydantic import BaseModel
import pytest


# ------------------------------------------------------------
# base tests
# ------------------------------------------------------------


def test_completion_defaults():
    response = completion({"role": "user", "content": "Hello, world!"})

    assert isinstance(response.choices[0].message.content, str)


def test_completion_string_input():
    response = completion("Hello, world!")

    assert isinstance(response.choices[0].message.content, str)


def test_completion_stream():
    response = completion("Repeat after me: I am a happy person", stream=True)

    content = ""

    for chunk in response:
        content += chunk.choices[0].delta.content or ""

    assert "happy" in content.lower()


@pytest.mark.asyncio
async def test_completion_async():
    response = await async_completion("Hello, world!")

    assert isinstance(response.choices[0].message.content, str)


# ------------------------------------------------------------
# context tests
# ------------------------------------------------------------


def test_completion_context():
    context = "My favorite color is blue."

    response = completion(
        "What is my favorite color?",
        context=context,
    )

    assert "blue" in response.choices[0].message.content


# ------------------------------------------------------------
# structured output tests
# ------------------------------------------------------------


def test_completion_structured_output_with_pydantic_response_model():
    class User(BaseModel):
        name: str
        age: int

    response = completion("Extract john is 20 years old", response_model=User)

    assert isinstance(response, User)

    assert isinstance(response.name, str)
    assert isinstance(response.age, int)


# using string response model
def test_completion_structured_output_with_string_response_model():
    response = completion(
        "Extract john is 20 years old", response_model=["name", "age : int"]
    )

    assert isinstance(response.name, str)
    assert isinstance(response.age, int)


# using type hint response model
def test_completion_structured_output_with_type_hint_response_model():
    response = completion("Extract john is 20 years old", response_model=int)

    assert isinstance(response, int)
    assert response == 20


# test with `response_format`
def test_completion_structured_output_with_response_format():
    class User(BaseModel):
        name: str
        age: int

    response = completion("Extract john is 20 years old", response_format=User)

    assert isinstance(response, User)

    assert isinstance(response.name, str)
    assert isinstance(response.age, int)


def test_completion_structured_output_with_instructor_mode():
    class User(BaseModel):
        name: str
        age: int

    response = completion(
        "Extract john is 20 years old",
        model="anthropic/claude-3-5-haiku-latest",
        response_model=User,
        instructor_mode="markdown_json_mode",
    )

    assert isinstance(response, User)

    assert isinstance(response.name, str)
    assert isinstance(response.age, int)


# ------------------------------------------------------------
# tool use tests
# ------------------------------------------------------------


# Test tool calling with openai function
def test_completion_with_openai_function():
    tool = {
        "type": "function",
        "function": {
            "name": "book_flight",
            "strict": True,
            "parameters": {
                "properties": {
                    "destination": {"title": "Destination", "type": "string"},
                    "return_date": {"title": "Return Date", "type": "string"},
                },
                "required": ["destination", "return_date"],
                "title": "book_flight",
                "type": "object",
                "additionalProperties": False,
            },
        },
    }

    response = completion(
        "Book a flight to Tokyo for return on 2025-01-01", tools=[tool]
    )

    assert response.choices[0].message.tool_calls[0].function.name == "book_flight"


def test_completion_with_function_tool():
    def book_flight(destination: str, return_date: str):
        return f"Booking a flight to {destination} for return on {return_date}"

    response = completion(
        "Book a flight to Tokyo for return on 2025-01-01", tools=[book_flight]
    )

    assert response.choices[0].message.tool_calls[0].function.name == "book_flight"


def test_completion_with_pydantic_model_tool():
    class BookFlight(BaseModel):
        destination: str
        return_date: str

    response = completion(
        "Book a flight to Tokyo for return on 2025-01-01", tools=[BookFlight]
    )

    assert response.choices[0].message.tool_calls[0].function.name == "BookFlight"


def test_completion_with_tool_execution():
    def book_flight(destination: str, return_date: str):
        return f"Booking a flight to {destination} for return on {return_date}"

    response = completion(
        "Book a flight to Tokyo for return on 2025-01-01",
        tools=[book_flight],
        run_tools=True,
    )

    assert response.choices[0].message.content != None


def test_completion_with_tool_execution_and_returned_messages():
    def book_flight(destination: str, return_date: str):
        return f"Booking a flight to {destination} for return on {return_date}"

    response = completion(
        "Book a flight to Tokyo for return on 2025-01-01",
        tools=[book_flight],
        run_tools=True,
        return_messages=True,
    )

    assert isinstance(response, list)
    assert len(response) == 3


# ------------------------------------------------------------
# batch completions tests
# ------------------------------------------------------------


# single model, multiple thread batch job
def test_single_model_batch_job():
    messages = [
        [{"role": "user", "content": "How are you?"}],
        [
            {"role": "system", "content": "The user's favorite color is green."},
            {"role": "user", "content": "What is my favorite color?"},
        ],
    ]

    response = completion(messages=messages, model="openai/gpt-4o-mini")

    assert len(response) == 2
    assert isinstance(response[0].model_dump(), dict)
    assert isinstance(response[1].model_dump(), dict)


# test mutiple model batch job
def test_multiple_model_batch_job():
    models = ["openai/gpt-4o", "openai/gpt-4o-mini"]

    response = completion({"role": "user", "content": "hi"}, model=models)

    assert len(response) == 2
    assert isinstance(response[0].model_dump(), dict)
    assert isinstance(response[1].model_dump(), dict)


# test batch structured outputs
def test_batch_job_structured_outputs():
    class User(BaseModel):
        name: str
        age: int

    response = completion(
        [
            [{"role": "user", "content": "Extract john is 20 years old"}],
            [{"role": "user", "content": "Extract jane is 30 years old"}],
        ],
        response_model=User,
    )

    assert isinstance(response[0].model_dump(), dict)
    assert isinstance(response[1].model_dump(), dict)

    assert response[0].name.lower() == "john"
    assert response[1].name.lower() == "jane"


# multiple model batch job with structured outputs
def test_multiple_model_batch_job_structured_outputs():
    class User(BaseModel):
        name: str
        age: int

    models = ["openai/gpt-4o", "openai/gpt-4o-mini"]

    response = completion(
        "extract john is 20 years old", model=models, response_model=User
    )

    assert len(response) == 2
    assert isinstance(response[0].model_dump(), dict)
    assert isinstance(response[1].model_dump(), dict)

    assert response[0].name.lower() == "john"
    assert response[1].name.lower() == "john"



import random
from typing import Optional


def get_random_name():
    names = [
        "John",
        "Alice",
        "Bob",
        "Grace",
        "David",
        "Eve",
        "Frank",
        "Hannah",
        "Hank",
        "Ivy",
        "Jack",
        "Jasmine",
        "Larry",
        "Katherine",
        "Noah",
        "Mabel",
        "Ryan",
        "Nina",
        "Victor",
        "Olivia",
        "Xavier",
        "Pam",
        "Zach",
        "Queenie",
        "Charlie",
        "Abigail",
        "Trent",
        "Beatrice",
        "Sasha",
        "Charlotte",
        "Quinn",
        "Diana",
        "Yvonne",
        "Eleanor",
        "Wendy",
        "Fiona",
        "Ursula",
        "Giselle",
        "Sasha",
        "Iris",
        "Tamsin",
        "Ophelia",
        "Vera",
        "Petunia",
        "Roxanne",
        "Selene",
    ]
    return random.choice(names)


# -----------------------------------------------------------------
# INSTRUCTION HELPER FUNCTIONS
# -----------------------------------------------------------------


def build_instruction(
    name: str,
    role: str,
    tool_names: Optional[list] = None,
    instructions: Optional[str] = None,
) -> dict:
    tool_string = ""
    intro_string = (
        f"You are {name}, a world class {role}."
        "As a genius expert, you are able to solve complex problems and provide insightful answers to any question, with a specialization using your role.\n\n"
    )

    if tool_names:
        tool_string = (
            f"You have access to the following tools:\n{tool_names}\n"
            "Use your given tools to assist you in completing your various tasks.\n\n"
        )

    return {
        "role": "system",
        "content": intro_string
        + (tool_string if tool_string else "")
        + (instructions if instructions else ""),
    }



from pathlib import Path
from typing import Any, Dict, Union, Optional
import yaml
from pydantic import BaseModel, create_model

from ...lib import XNANOException, console


def convert_yaml_to_pydantic(
    path: Union[str, Path], verbose: bool = False
) -> BaseModel:
    """
    Reads a YAML file and creates a Pydantic model from its contents.

    Args:
        path (Union[str, Path]): The path to the YAML file.

    Returns:
        BaseModel: An instance of a Pydantic model representing the YAML structure.
    """
    if isinstance(path, str):
        path = Path(path)

    if verbose:
        console.message(f"Converting YAML to Pydantic model from {path}")

    try:
        with open(path, "r", encoding="utf-8") as file:
            yaml_content = yaml.safe_load(file)
            if verbose:
                console.message(f"YAML content loaded: {yaml_content}")
            model_class = create_pydantic_model(yaml_content, verbose=verbose)
            return model_class(
                **yaml_content
            )  # Instantiate the model with the loaded data
    except Exception as e:
        raise XNANOException(f"Error reading YAML {path}: {str(e)}")


def create_pydantic_model(
    data: Dict[str, Any], model_name: str = "DynamicModel", verbose: bool = False
) -> BaseModel:
    """
    Recursively creates a Pydantic model from a dictionary.

    Args:
        data (Dict[str, Any]): The dictionary representation of the YAML content.
        model_name (str): The name of the Pydantic model to create.

    Returns:
        BaseModel: A Pydantic model representing the data structure.
    """
    fields = {}

    for key, value in data.items():
        if isinstance(value, dict):
            # Create a nested model for dictionaries without the DynamicModel_ prefix
            nested_model = create_pydantic_model(value, model_name=key, verbose=verbose)
            fields[key] = (nested_model, ...)  # Add nested model as a field
        else:
            # Handle primitive values
            if value is None:
                fields[key] = (Optional[Any], None)  # Treat null as Optional
            elif isinstance(value, str):
                fields[key] = (str, value)  # Use string type with actual value
            elif isinstance(value, int):
                fields[key] = (int, value)  # Use int type with actual value
            elif isinstance(value, float):
                fields[key] = (float, value)  # Use float type with actual value
            elif isinstance(value, bool):
                fields[key] = (bool, value)  # Use bool type with actual value
            else:
                fields[key] = (Any, value)  # Fallback to Any for other types

    model = create_model(model_name, **fields)
    return model


if __name__ == "__main__":
    model_instance = convert_yaml_to_pydantic("config.yaml", verbose=True)
    print(model_instance)
    print(model_instance.model_dump())



# tool calling helpers & converters


from pydantic import create_model, Field, BaseModel
import inspect
from typing import Any, Callable, Dict, Type, Union

from ....lib import XNANOException
from ....types.completions.tools import Tool


# tool generator
def generate_tool(client, tool_name: str, model: str) -> Tool:
    """Dynamically generate a tool function from its name"""

    function_prompt = f"""Create a Python function named '{tool_name}' that:
    
    1. Has typed arguments and return value
    2. Includes clear docstring
    3. Handles errors with try/except
    4. Places imports inside function body
    5. Implements core functionality implied by '{tool_name}'
    6. OPTS TO USING STANDARD PYTHON LIBRARIES UNLESS SPECIFICALLY REQUESTED
    """

    from ...generators.code_generators import generate_code

    # Use coder to generate the implementation
    generated_code = generate_code(
        instructions=function_prompt,
        model=model,
        temperature=0.2,
        client=client,
        verbose=client.verbose,
        return_code=True,
        progress_bar=False,
    )

    # Create function object from code
    namespace = {}
    try:
        exec(generated_code, namespace)
        generated_function = namespace[tool_name]
    except Exception as e:
        raise ValueError(f"Failed to create function object for {tool_name}: {str(e)}")

    # Create and return a Tool instance
    tool = Tool(
        name=get_function_name(generated_function),
        function=generated_function,
        description=get_function_description(generated_function),
        arguments=get_function_arguments(generated_function),
        formatted_function=convert_to_openai_tool(generated_function),
    )

    return tool


# convert to tool model
def convert_to_tool(function: Union[Callable, Type[BaseModel], Dict[str, Any]]) -> Tool:
    """
    Builds a raw function/class into a CompletionsTool.

    Args:
        function (Union[Callable, Type[BaseModel], Dict[str, Any]]): The function/class to convert.

    Returns:
        CompletionsTool: The CompletionsTool.
    """

    # setup tool
    tool = Tool(function=function)

    try:
        # format tool
        tool.formatted_function = convert_to_openai_tool(function)
    except Exception as e:
        raise XNANOException(f"Failed to convert tool: {e}")

    try:
        tool.name = get_function_name(function)
        tool.description = get_function_description(function)
        tool.arguments = get_function_arguments(function)
    except Exception as e:
        raise XNANOException(f"Failed to format tool: {e}")

    return tool


# convert to openai tool
def convert_to_openai_tool(
    function: Union[Dict[str, Any], Type[BaseModel], Callable],
) -> Dict[str, Any]:
    """Convert a raw function/class to an OpenAI tool.

    Examples:
        >>> convert_to_openai_tool(lambda x: x + 1)
        >>> convert_to_openai_tool(MyModel)
        >>> convert_to_openai_tool({"name": "my_tool", "description": "my tool", "function": lambda x: x + 1})

    Args:
        function (Union[Dict[str, Any], Type[BaseModel], Callable]): The function/class to convert.

    Returns:
        Dict[str, Any]: The OpenAI tool.
    """
    from openai import pydantic_function_tool

    if isinstance(function, dict):
        # If it's already a dictionary, assume it's in the correct format
        return function
    elif isinstance(function, type) and issubclass(function, BaseModel):
        # If it's a Pydantic model, use pydantic_function_tool directly
        return pydantic_function_tool(function)
    elif callable(function):
        # If it's a callable, convert it to a Pydantic model first
        pydantic_model = create_pydantic_model_from_function(function)
        return pydantic_function_tool(pydantic_model)
    else:
        raise ValueError(f"Unsupported function type: {type(function)}")


# create pydantic model from function
def create_pydantic_model_from_function(function: Callable) -> Type[BaseModel]:
    """Create a Pydantic model from a Python function.

    Args:
        function (Callable): The function to convert.

    Returns:
        Type[BaseModel]: The Pydantic model.
    """
    signature = inspect.signature(function)
    fields = {}
    for name, param in signature.parameters.items():
        annotation = (
            param.annotation if param.annotation != inspect.Parameter.empty else Any
        )
        default = ... if param.default == inspect.Parameter.empty else param.default
        fields[name] = (annotation, Field(default=default))

    # Use the original function name for the model
    model = create_model(function.__name__, **fields)
    model.__doc__ = function.__doc__ or ""

    # Store the original function as an attribute
    model._original_function = function

    return model


# get function name helper
def get_function_name(
    function: Union[Callable, Type[BaseModel], Dict[str, Any]],
) -> str:
    """Get the name of a function."""

    # if in callable format
    if callable(function):
        return function.__name__
    # if in model format
    elif isinstance(function, type) and issubclass(function, BaseModel):
        return function.__name__
    # if already openai dict
    elif isinstance(function, dict):
        return function.get("name", None)
    else:
        raise ValueError(f"Unsupported function type: {type(function)}")


# get function description helper
def get_function_description(
    function: Union[Callable, Type[BaseModel], Dict[str, Any]],
) -> str:
    """Get the description of a function

    Args:
        function (Union[Callable, Type[BaseModel], Dict[str, Any]]): The function to get the description of.

    Returns:
        str: The description of the function.
    """
    if isinstance(function, dict):
        return function.get("description", None)
    else:
        return function.__doc__


# get function arguments helper
def get_function_arguments(
    function: Union[Callable, Type[BaseModel], Dict[str, Any]],
) -> Dict[str, Any]:
    """Get the arguments of a function.

    Args:
        function (Union[Callable, Type[BaseModel], Dict[str, Any]]): The function to get the arguments of.

    Returns:
        Dict[str, Any]: The arguments of the function.
    """
    # if in callable format
    if callable(function):
        return inspect.signature(function).parameters
    # if in model format
    elif isinstance(function, type) and issubclass(function, BaseModel):
        return function.model_fields
    elif isinstance(function, dict):
        if function.get("type", None) == "function":
            return function.get("parameters", {}).get("properties", {})
        else:
            return {}
    else:
        raise ValueError(f"Unsupported function type: {type(function)}")


if __name__ == "__main__":
    # Example usage of get_function_name
    def example_function(x):
        return x + 1

    class ExampleModel(BaseModel):
        field: int

    example_dict = {
        "name": "example_dict",
        "description": "This is an example dictionary",
    }

    print(get_function_name(example_function))  # Output: example_function
    print(get_function_name(ExampleModel))  # Output: ExampleModel
    print(get_function_name(example_dict))  # Output: example_dict

    # Example usage of get_function_description
    def example_function_with_doc(x):
        """This is an example function with a docstring."""
        return x + 1

    example_dict_with_description = {
        "name": "example_dict",
        "description": "This is an example dictionary with a description",
    }

    print(
        get_function_description(example_function_with_doc)
    )  # Output: This is an example function with a docstring.
    print(
        get_function_description(ExampleModel)
    )  # Output: None (assuming BaseModel does not have a docstring)
    print(
        get_function_description(example_dict_with_description)
    )  # Output: This is an example dictionary with a description

    # Example usage of get_function_arguments
    def example_function_with_args(a, b, c=3):
        return a + b + c

    class ExampleModelWithFields(BaseModel):
        field1: int
        field2: str

    print(
        get_function_arguments(example_function_with_args)
    )  # Output: OrderedDict([('a', <Parameter "a">), ('b', <Parameter "b">), ('c', <Parameter "c=3">)])
    print(
        get_function_arguments(ExampleModelWithFields)
    )  # Output: {'field1': <FieldInfo>, 'field2': <FieldInfo>}



__all__ = [
    "generate_code",
    "generate_function",
    "generate_classification",
    "async_generate_classification",
    "generate_chunks",
    "generate_extraction",
    "async_generate_extraction",
    "generate_sql",
    "generate_system_prompt",
    "generate_qa_pairs",
    "generate_answers",
    "generate_questions",
    "generate_validation",
    "async_generate_validation",
    # multimodal
    "multimodal_generate_image",
    "multimodal_generate_audio",
    "multimodal_generate_transcription",
]


from ...lib.router import router


class generate_code(router):
    pass


generate_code.init("xnano.resources.generators.code_generators", "generate_code")


class generate_function(router):
    pass


generate_function.init(
    "xnano.resources.generators.code_generators", "generate_function"
)


class generate_classification(router):
    pass


generate_classification.init(
    "xnano.resources.generators.classifier", "generate_classification"
)


class async_generate_classification(router):
    pass


async_generate_classification.init(
    "xnano.resources.generators.classifier", "async_generate_classification"
)


class generate_chunks(router):
    pass


generate_chunks.init("xnano.resources.generators.chunker", "generate_chunks")


class generate_extraction(router):
    pass


generate_extraction.init("xnano.resources.generators.extractor", "generate_extraction")


class async_generate_extraction(router):
    pass


async_generate_extraction.init(
    "xnano.resources.generators.extractor", "async_generate_extraction"
)


class generate_sql(router):
    pass


generate_sql.init("xnano.resources.generators.generate_sql", "generate_sql")


class generate_system_prompt(router):
    pass


generate_system_prompt.init(
    "xnano.resources.generators.prompting", "generate_system_prompt"
)


class generate_qa_pairs(router):
    pass


generate_qa_pairs.init(
    "xnano.resources.generators.question_answer", "generate_qa_pairs"
)


class generate_answers(router):
    pass


generate_answers.init("xnano.resources.generators.question_answer", "generate_answers")


class generate_questions(router):
    pass


generate_questions.init(
    "xnano.resources.generators.question_answer", "generate_questions"
)


class generate_validation(router):
    pass


generate_validation.init("xnano.resources.generators.validator", "generate_validation")


class async_generate_validation(router):
    pass


async_generate_validation.init(
    "xnano.resources.generators.validator", "async_generate_validation"
)


class multimodal_generate_image(router):
    pass


multimodal_generate_image.init(
    "xnano.resources.generators.multimodal", "multimodal_generate_image"
)


class multimodal_generate_audio(router):
    pass


multimodal_generate_audio.init(
    "xnano.resources.generators.multimodal", "multimodal_generate_audio"
)


class multimodal_generate_transcription(router):
    pass


multimodal_generate_transcription.init(
    "xnano.resources.generators.multimodal", "multimodal_generate_transcription"
)



__all__ = ["text_reader", "text_embeddings", "text_chunker"]


from ...lib.router import router


class text_reader(router):
    pass


text_reader.init("xnano.resources.text.documents.text_reader", "text_reader")


class text_embeddings(router):
    pass


text_embeddings.init(
    "xnano.resources.text.embeddings.text_embeddings", "text_embeddings"
)


class text_chunker(router):
    pass


text_chunker.init("xnano.resources.text.processing.text_chunker", "text_chunker")



# full test suite for xnano

from tests.test_agents import (
    test_agents_blank_init,
    test_agents_completion_with_short_term_memory,
    test_agents_single_completion_context,
)
from tests.test_completion import (
    test_completion_defaults,
    test_completion_context,
    test_completion_string_input,
    test_completion_stream,
    test_completion_async,
    test_completion_structured_output_with_pydantic_response_model,
    test_completion_structured_output_with_instructor_mode,
    test_completion_structured_output_with_response_format,
    test_completion_structured_output_with_string_response_model,
    test_completion_structured_output_with_type_hint_response_model,
    test_completion_with_tool_execution,
    test_completion_with_tool_execution_and_returned_messages,
    test_completion_with_function_tool,
    test_completion_with_openai_function,
    test_completion_with_pydantic_model_tool,
    test_single_model_batch_job,
    test_batch_job_structured_outputs,
    test_multiple_model_batch_job,
    test_multiple_model_batch_job_structured_outputs,
)
from tests.test_generators import (
    test_generators_chunk_generation,
    test_generators_code_generation,
    test_generators_function_generator,
    test_generators_multi_label_classification,
    test_generators_single_label_classification,
    test_generators_create_qa_pairs,
    test_generators_extraction,
    test_generators_extraction_batch_inputs,
    test_generators_sql_query,
    test_generators_system_prompt,
    test_generators_validator_guardrails,
)
from tests.test_generative_model import (
    test_generative_model_init,
    test_generative_model_patch,
    test_generative_model_completion,
    test_generative_model_generate_single,
    test_generative_model_generate_multiple,
    test_generative_model_generate_single_with_messages,
    test_generative_model_field_regeneration,
)


if __name__ == "__main__":
    import pytest

    pytest.main()



# xnano . models

__all__ = ["GenerativeModel", "model_patch", "model_unpatch"]

# primary import level for xnano.GenerativeModel
# (Pydantic Subclass & Extension)

from .resources.models.mixin import (
    GenerativeModel,
    # TODO:
    # update lib imports to use .model_ at
    # internal level
    patch as model_patch,
    unpatch as model_unpatch,
)



from pydantic import BaseModel
from typing import Any, Callable, Type, Dict, Optional, Union, Literal


# tool type
ToolType = Union[str, Callable, callable, Type[BaseModel], Dict[str, Any]]


# tool choice parameter
ToolChoice = Literal["auto", "none", "required"]


# tool base
class Tool(BaseModel):
    """Internal tool class."""

    class Config:
        arbitrary_types_allowed = True

    # tool name (used for execution)
    name: Optional[str] = None

    # arguments (used for execution)
    arguments: Optional[Union[Dict[str, Any], Dict, Any]] = None

    # description
    description: Optional[str] = None

    # function
    function: Union[
        # openai function
        Dict[str, Any],
        # pydantic model
        Type[BaseModel],
        # callable
        Callable,
    ]

    # formatted function (completion request)
    formatted_function: Optional[Dict[str, Any]] = None

    is_string_tool: Optional[bool] = False



# xnano . generators

__all__ = [
    "generate_code",
    "generate_function",
    "generate_classification",
    "async_generate_classification",
    "generate_chunks",
    "generate_extraction",
    "async_generate_extraction",
    "generate_sql",
    "generate_system_prompt",
    "generate_qa_pairs",
    "generate_answers",
    "generate_questions",
    "generate_validation",
    "async_generate_validation",
    # multimodal
    "multimodal_generate_image",
    "multimodal_generate_audio",
    "multimodal_generate_transcription",
]

# xnano follows a generate_ namespace, to help
# identify functions that use LLMS to make calls
# this is essentially so things like a confusion of

# .text_chunker() & .generate_chunks() are easily distinguishable
# as one not being an llm method

from .resources.generators._routing import (
    generate_code,
    generate_function,
    generate_classification,
    async_generate_classification,
    generate_chunks,
    generate_extraction,
    async_generate_extraction,
    generate_sql,
    generate_system_prompt,
    generate_qa_pairs,
    generate_answers,
    generate_questions,
    generate_validation,
    async_generate_validation,
    # multimodal
    multimodal_generate_image,
    multimodal_generate_audio,
    multimodal_generate_transcription,
)



# memory types

from enum import Enum
from typing import Literal, Union, Type, List, Dict, Any
from ..data.document import Document


# memory location literal helper
MemoryLocationParam = Union[Literal[":memory:"], str]


# distance types
MemoryDistanceParam = Literal["cosine", "euclid", "dot", "manhattan"]


# data types
MemoryDataTypeParam = Union[
    str, List[str], Dict[str, Any], List[Dict[str, Any]], Document, List[Document]
]



__all__ = [
    "VectorStore",
    "Database",
    "web_reader",
    "web_search",
    "web_news_search",
    "async_web_search",
    "async_web_news_search",
    "async_web_reader",
]


from ...lib.router import router


class VectorStore(router):
    pass


VectorStore.init("xnano.resources.data.vector_store", "VectorStore")


class Database(router):
    pass


Database.init("xnano.resources.data.database", "Database")


class web_reader(router):
    pass


web_reader.init("xnano.resources.data.web", "web_reader")


class web_search(router):
    pass


web_search.init("xnano.resources.data.web", "web_search")


class web_news_search(router):
    pass


web_news_search.init("xnano.resources.data.web", "web_news_search")


class async_web_search(router):
    pass


async_web_search.init("xnano.resources.data.web", "async_web_search")


class async_web_news_search(router):
    pass


async_web_news_search.init("xnano.resources.data.web", "async_web_news_search")


class async_web_reader(router):
    pass


async_web_reader.init("xnano.resources.data.web", "async_web_reader")



__all__ = [
    "convert_yaml_to_pydantic",
    "format_messages",
    "repair_messages",
    "verify_messages_integrity",
    "swap_system_prompt",
    "convert_to_openai_tool",
    "MessageUtils",
    "Tools",
]


from ...lib.router import router


class convert_yaml_to_pydantic(router):
    pass


convert_yaml_to_pydantic.init(
    "xnano.resources.utils.convert_yaml_to_pydantic", "convert_yaml_to_pydantic"
)


class format_messages(router):
    pass


format_messages.init("xnano.resources.utils.messages", "format_messages")


class repair_messages(router):
    pass


repair_messages.init("xnano.resources.utils.messages", "repair_messages")


class verify_messages_integrity(router):
    pass


class swap_system_prompt(router):
    pass


swap_system_prompt.init("xnano.resources.utils.messages", "swap_system_prompt")


class convert_to_openai_tool(router):
    pass


convert_to_openai_tool.init(
    "xnano.resources.completions.resources.tool_calling", "convert_to_openai_tool"
)


class MessageUtils(router):
    pass


MessageUtils.init("xnano.resources.utils.messages", "Messages")


class Tools(router):
    pass


Tools.init("xnano.resources.utils.tools", "Tools")



# openai types

from openai.types.chat.chat_completion_audio_param import ChatCompletionAudioParam
from openai.types.chat.chat_completion_modality import ChatCompletionModality
from openai.types.chat.chat_completion_prediction_content_param import (
    ChatCompletionPredictionContentParam,
)
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai._client import Stream, AsyncStream
